# Copyright 2023-2025 Broadcom.
# SPDX-License-Identifier: Apache-2
"""
Common classes and values used around relenv.
"""
import http.client
import logging
import os
import pathlib
import platform
import queue
import selectors
import subprocess
import sys
import tarfile
import textwrap
import threading
import time

# relenv package version
__version__ = "0.20.6"

MODULE_DIR = pathlib.Path(__file__).resolve().parent

DEFAULT_PYTHON = "3.10.18"

LINUX = "linux"
WIN32 = "win32"
DARWIN = "darwin"

MACOS_DEVELOPMENT_TARGET = "10.15"

REQUEST_HEADERS = {"User-Agent": f"relenv {__version__}"}

CHECK_HOSTS = (
    "packages.broadcom.com/artifactory/saltproject-generic",
    "woz.io",
)

arches = {
    LINUX: (
        "x86_64",
        "aarch64",
    ),
    DARWIN: ("x86_64", "arm64"),
    WIN32: (
        "amd64",
        "x86",
        #    "arm64", # Python 11 should support arm.
    ),
}


if sys.platform == "win32":
    DEFAULT_DATA_DIR = pathlib.Path.home() / "AppData" / "Local" / "relenv"
else:
    DEFAULT_DATA_DIR = pathlib.Path.home() / ".local" / "relenv"

DATA_DIR = pathlib.Path(os.environ.get("RELENV_DATA", DEFAULT_DATA_DIR)).resolve()

SHEBANG_TPL_LINUX = textwrap.dedent(
    """#!/bin/sh
"true" ''''
"exec" "$(dirname "$(readlink -f "$0")"){}" "$0" "$@"
'''
"""
)

SHEBANG_TPL_MACOS = textwrap.dedent(
    """\
#!/bin/sh
"true" ''''
TARGET_FILE=$0
cd "$(dirname "$TARGET_FILE")" || return
TARGET_FILE=$(basename "$TARGET_FILE")
# Iterate down a (possible) chain of symlinks
while [ -L "$TARGET_FILE" ]
do
    TARGET_FILE=$(readlink "$TARGET_FILE")
    cd "$(dirname "$TARGET_FILE")" || return
    TARGET_FILE=$(basename "$TARGET_FILE")
done
PHYS_DIR=$(pwd -P)
REALPATH=$PHYS_DIR/$TARGET_FILE
"exec" "$(dirname "$REALPATH")"{} "$REALPATH" "$@"
'''"""
)

if sys.platform == "linux":
    SHEBANG_TPL = SHEBANG_TPL_LINUX
else:
    SHEBANG_TPL = SHEBANG_TPL_MACOS


log = logging.getLogger(__name__)


class RelenvException(Exception):
    """
    Base class for exeptions generated from relenv.
    """


def format_shebang(python, tpl=SHEBANG_TPL):
    """
    Return a formatted shebang.
    """
    return tpl.format(python).strip()


def build_arch():
    """
    Return the current machine.
    """
    machine = platform.machine()
    return machine.lower()


def work_root(root=None):
    """
    Get the root directory that all other relenv working directories should be based on.

    :param root: An explicitly requested root directory
    :type root: str

    :return: An absolute path to the relenv root working directory
    :rtype: ``pathlib.Path``
    """
    if root is not None:
        base = pathlib.Path(root).resolve()
    else:
        base = MODULE_DIR
    return base


def work_dir(name, root=None):
    """
    Get the absolute path to the relenv working directory of the given name.

    :param name: The name of the directory
    :type name: str
    :param root: The root directory that this working directory will be relative to
    :type root: str

    :return: An absolute path to the requested relenv working directory
    :rtype: ``pathlib.Path``
    """
    root = work_root(root)
    if root == MODULE_DIR:
        base = root / "_{}".format(name)
    else:
        base = root / name
    return base


class WorkDirs:
    """
    Simple class used to hold references to working directories relenv uses relative to a given root.

    :param root: The root of the working directories tree
    :type root: str
    """

    def __init__(self, root):
        self.root = root
        self.data = DATA_DIR
        self.toolchain_config = work_dir("toolchain", self.root)
        self.toolchain = work_dir("toolchain", DATA_DIR)
        self.build = work_dir("build", DATA_DIR)
        self.src = work_dir("src", DATA_DIR)
        self.logs = work_dir("logs", DATA_DIR)
        self.download = work_dir("download", DATA_DIR)

    def __getstate__(self):
        """
        Return an object used for pickling.

        :return: The picklable state
        """
        return {
            "root": self.root,
            "toolchain_config": self.toolchain_config,
            "toolchain": self.toolchain,
            "build": self.build,
            "src": self.src,
            "logs": self.logs,
            "download": self.download,
        }

    def __setstate__(self, state):
        """
        Unwrap the object returned from unpickling.

        :param state: The state to unpickle
        :type state: dict
        """
        self.root = state["root"]
        self.toolchain_config = state["toolchain_config"]
        self.toolchain = state["toolchain"]
        self.build = state["build"]
        self.src = state["src"]
        self.logs = state["logs"]
        self.download = state["download"]


def work_dirs(root=None):
    """
    Returns a WorkDirs instance based on the given root.

    :param root: The desired root of relenv's working directories
    :type root: str

    :return: A WorkDirs instance based on the given root
    :rtype: ``relenv.common.WorkDirs``
    """
    return WorkDirs(work_root(root))


def get_toolchain(arch=None, root=None):
    """
    Get a the toolchain directory, specific to the arch if supplied.

    :param arch: The architecture to get the toolchain for
    :type arch: str
    :param root: The root of the relenv working directories to search in
    :type root: str

    :return: The directory holding the toolchain
    :rtype: ``pathlib.Path``
    """
    os.makedirs(DATA_DIR, exist_ok=True)
    if sys.platform != "linux":
        return DATA_DIR

    TOOLCHAIN_ROOT = DATA_DIR / "toolchain"
    TOOLCHAIN_PATH = TOOLCHAIN_ROOT / get_triplet()
    if TOOLCHAIN_PATH.exists():
        return TOOLCHAIN_PATH

    ppbt = None

    try:
        import ppbt.common
    except ImportError:
        pass

    if ppbt:
        TOOLCHAIN_ROOT.mkdir(exist_ok=True)
        ppbt.common.extract_archive(str(TOOLCHAIN_ROOT), str(ppbt.common.ARCHIVE))
        return TOOLCHAIN_PATH


def get_triplet(machine=None, plat=None):
    """
    Get the target triplet for the specified machine and platform.

    If any of the args are None, it will try to deduce what they should be.

    :param machine: The machine for the triplet
    :type machine: str
    :param plat: The platform for the triplet
    :type plat: str

    :raises RelenvException: If the platform is unknown

    :return: The target triplet
    :rtype: str
    """
    if not plat:
        plat = sys.platform
    if not machine:
        machine = build_arch()
    if plat == "darwin":
        return f"{machine}-macos"
    elif plat == "win32":
        return f"{machine}-win"
    elif plat == "linux":
        return f"{machine}-linux-gnu"
    else:
        raise RelenvException(f"Unknown platform {plat}")


def plat_from_triplet(plat):
    """
    Convert platform from build to the value of sys.platform.
    """
    if plat == "linux-gnu":
        return "linux"
    elif plat == "macos":
        return "darwin"
    elif plat == "win":
        return "win32"
    raise RelenvException(f"Unkown platform {plat}")


def list_archived_builds():
    """
    Return a list of version, architecture and platforms for builds.
    """
    builds = []
    dirs = work_dirs(DATA_DIR)
    for root, dirs, files in os.walk(dirs.build):
        for file in files:
            if file.endswith(".tar.xz"):
                file = file[:-7]
                version, triplet = file.split("-", 1)
                arch, plat = triplet.split("-", 1)
                builds.append((version, arch, plat))
    return builds


def archived_build(triplet=None):
    """
    Finds a the location of an archived build.

    :param triplet: The build triplet to find
    :type triplet: str

    :return: The location of the archived build
    :rtype: ``pathlib.Path``
    """
    if not triplet:
        triplet = get_triplet()
    dirs = work_dirs(DATA_DIR)
    archive = f"{triplet}.tar.xz"
    return dirs.build / archive


def extract_archive(to_dir, archive):
    """
    Extract an archive to a specific location.

    :param to_dir: The directory to extract to
    :type to_dir: str
    :param archive: The archive to extract
    :type archive: str
    """
    if archive.endswith("tgz"):
        read_type = "r:gz"
    elif archive.endswith("xz"):
        read_type = "r:xz"
    elif archive.endswith("bz2"):
        read_type = "r:bz2"
    else:
        read_type = "r"
    with tarfile.open(archive, read_type) as t:
        t.extractall(to_dir)


def get_download_location(url, dest):
    """
    Get the full path to where the url will be downloaded to.

    :param url: The url to donwload
    :type url: str
    :param dest: Where to download the url to
    :type dest: str

    :return: The path to where the url will be downloaded to
    :rtype: str
    """
    return os.path.join(dest, os.path.basename(url))


def check_url(url, timestamp=None, timeout=30):
    """
    Check that the url returns a 200.
    """
    # Late import so we do not import hashlib before runtime.bootstrap is called.
    import time
    import urllib.request

    headers = dict(REQUEST_HEADERS)
    req = urllib.request.Request(url)

    if timestamp:
        headers["If-Modified-Since"] = time.strftime(
            "%a, %d %b %Y %H:%M:%S GMT", time.gmtime(timestamp)
        )

    for k, v in headers.items():
        req.add_header(k, v)

    fin = None
    try:
        fin = urllib.request.urlopen(req, timeout=timeout)
    except Exception as exc:
        print(exc)
        return False
    finally:
        if fin:
            fin.close()
    return True


def fetch_url(url, fp, backoff=3, timeout=30):
    """
    Fetch the contents of a url.

    This method will store the contents in the given file like object.
    """
    # Late import so we do not import hashlib before runtime.bootstrap is called.
    import urllib.error
    import urllib.request

    last = time.time()
    if backoff < 1:
        backoff = 1
    n = 0
    while n < backoff:
        n += 1
        try:
            fin = urllib.request.urlopen(url, timeout=timeout)
        except (
            urllib.error.HTTPError,
            urllib.error.URLError,
            http.client.RemoteDisconnected,
        ) as exc:
            if n >= backoff:
                raise RelenvException(f"Error fetching url {url} {exc}")
            log.debug("Unable to connect %s", url)
            time.sleep(n * 10)
    log.info("url opened %s", url)
    try:
        total = 0
        size = 1024 * 300
        block = fin.read(size)
        while block:
            total += size
            if time.time() - last > 10:
                log.info("%s > %d", url, total)
                last = time.time()
            fp.write(block)
            block = fin.read(10240)
    finally:
        fin.close()
        # fp.close()
    log.info("Download complete %s", url)


def fetch_url_content(url, backoff=3, timeout=30):
    """
    Fetch the contents of a url.

    This method will store the contents in the given file like object.
    """
    # Late import so we do not import hashlib before runtime.bootstrap is called.
    import gzip
    import io
    import urllib.error
    import urllib.request

    fp = io.BytesIO()

    last = time.time()
    if backoff < 1:
        backoff = 1
    n = 0
    while n < backoff:
        n += 1
        try:
            fin = urllib.request.urlopen(url, timeout=timeout)
        except (
            urllib.error.HTTPError,
            urllib.error.URLError,
            http.client.RemoteDisconnected,
        ) as exc:
            if n >= backoff:
                raise RelenvException(f"Error fetching url {url} {exc}")
            log.debug("Unable to connect %s", url)
            time.sleep(n * 10)
    log.info("url opened %s", url)
    try:
        total = 0
        size = 1024 * 300
        block = fin.read(size)
        while block:
            total += size
            if time.time() - last > 10:
                log.info("%s > %d", url, total)
                last = time.time()
            fp.write(block)
            block = fin.read(10240)
    finally:
        fin.close()
        # fp.close()
    log.info("Download complete %s", url)
    fp.seek(0)
    info = fin.info()
    if "content-encoding" in info:
        if info["content-encoding"] == "gzip":
            log.debug("Found gzipped content")
            fp = gzip.GzipFile(fileobj=fp)
    return fp.read().decode()


def download_url(url, dest, verbose=True, backoff=3, timeout=60):
    """
    Download the url to the provided destination.

    This method assumes the last part of the url is a filename. (https://foo.com/bar/myfile.tar.xz)

    :param url: The url to download
    :type url: str
    :param dest: Where to download the url to
    :type dest: str
    :param verbose: Print download url and destination to stdout
    :type verbose: bool

    :raises urllib.error.HTTPError: If the url was unable to be downloaded

    :return: The path to the downloaded content
    :rtype: str
    """
    local = get_download_location(url, dest)
    if verbose:
        print(f"Downloading {url} -> {local}")
    fout = open(local, "wb")
    try:
        fetch_url(url, fout, backoff, timeout)
    except Exception as exc:
        if verbose:
            print(f"Unable to download: {url} {exc}", file=sys.stderr, flush=True)
        try:
            os.unlink(local)
        except OSError:
            pass
        raise
    return local


def runcmd(*args, **kwargs):
    """
    Run a command.

    Run the provided command, raising an Exception when the command finishes
    with a non zero exit code.  Arguments are passed through to ``subprocess.run``

    :return: The process result
    :rtype: ``subprocess.CompletedProcess``

    :raises RelenvException: If the command finishes with a non zero exit code
    """
    log.debug("Running command: %s", " ".join(args[0]))
    # if "stdout" not in kwargs:
    kwargs["stdout"] = subprocess.PIPE
    # if "stderr" not in kwargs:
    kwargs["stderr"] = subprocess.PIPE
    if "universal_newlines" not in kwargs:
        kwargs["universal_newlines"] = True
    if sys.platform != "win32":

        p = subprocess.Popen(*args, **kwargs)
        # Read both stdout and stderr simultaneously
        sel = selectors.DefaultSelector()
        sel.register(p.stdout, selectors.EVENT_READ)
        sel.register(p.stderr, selectors.EVENT_READ)
        ok = True
        while ok:
            for key, val1 in sel.select():
                line = key.fileobj.readline()
                if not line:
                    ok = False
                    break
                if line.endswith("\n"):
                    line = line[:-1]
                if key.fileobj is p.stdout:
                    log.info(line)
                else:
                    log.error(line)

    else:

        def enqueue_stream(stream, queue, type):
            NOOP = object()
            for line in iter(stream.readline, NOOP):
                if line is NOOP or line == "":
                    break
                if line:
                    queue.put((type, line))
            log.debug("stream close %r %r", type, line)
            stream.close()

        def enqueue_process(process, queue):
            process.wait()
            queue.put(("x", "x"))

        p = subprocess.Popen(*args, **kwargs)
        q = queue.Queue()
        to = threading.Thread(target=enqueue_stream, args=(p.stdout, q, 1))
        te = threading.Thread(target=enqueue_stream, args=(p.stderr, q, 2))
        tp = threading.Thread(target=enqueue_process, args=(p, q))
        te.start()
        to.start()
        tp.start()

        while True:
            kind, line = q.get()
            if kind == 1:  # stdout
                log.info(line[:-1])
            elif kind == 2:
                log.error(line[:-1])
            elif kind == "x":
                log.debug("process queue end")
                break

        tp.join()
        to.join()
        te.join()

    p.wait()
    if p.returncode != 0:
        raise RelenvException("Build cmd '{}' failed".format(" ".join(args[0])))
    return p


def relative_interpreter(root_dir, scripts_dir, interpreter):
    """
    Return a relativized path to the given scripts_dir and interpreter.
    """
    scripts = pathlib.Path(scripts_dir)
    interp = pathlib.Path(interpreter)
    root = pathlib.Path(root_dir)
    try:
        relinterp = interp.relative_to(root)
    except ValueError:
        raise ValueError("interperter not relative to root_dir")
    try:
        relscripts = pathlib.Path(*(".." for x in scripts.relative_to(root).parts))
    except ValueError:
        raise ValueError("scripts_dir not relative to root_dir")
    return relscripts / relinterp


def makepath(*paths):
    """
    Make a normalized path name from paths.
    """
    dir = os.path.join(*paths)
    try:
        dir = os.path.abspath(dir)
    except OSError:
        pass
    return dir, os.path.normcase(dir)


def addpackage(sitedir, name):
    """
    Add editable package to path.
    """
    import io
    import stat

    fullname = os.path.join(sitedir, name)
    paths = []
    try:
        st = os.lstat(fullname)
    except OSError:
        return
    if (getattr(st, "st_flags", 0) & stat.UF_HIDDEN) or (
        getattr(st, "st_file_attributes", 0) & stat.FILE_ATTRIBUTE_HIDDEN
    ):
        # print(f"Skipping hidden .pth file: {fullname!r}")
        return
    # print(f"Processing .pth file: {fullname!r}")
    try:
        # locale encoding is not ideal especially on Windows. But we have used
        # it for a long time. setuptools uses the locale encoding too.
        f = io.TextIOWrapper(io.open_code(fullname), encoding="locale")
    except OSError:
        return
    with f:
        for n, line in enumerate(f):
            if line.startswith("#"):
                continue
            if line.strip() == "":
                continue
            try:
                if line.startswith(("import ", "import\t")):
                    exec(line)
                    continue
                line = line.rstrip()
                dir, dircase = makepath(sitedir, line)
                if dircase not in paths and os.path.exists(dir):
                    paths.append(dir)
            except Exception:
                print(
                    "Error processing line {:d} of {}:\n".format(n + 1, fullname),
                    file=sys.stderr,
                )
                import traceback

                for record in traceback.format_exception(*sys.exc_info()):
                    for line in record.splitlines():
                        print("  " + line, file=sys.stderr)
                print("\nRemainder of file ignored", file=sys.stderr)
                break
    return paths


def sanitize_sys_path(sys_path_entries):
    """
    Sanitize `sys.path` to only include paths relative to the onedir environment.
    """
    __sys_path = []
    __valid_path_prefixes = tuple(
        {
            pathlib.Path(sys.prefix).resolve(),
            pathlib.Path(sys.base_prefix).resolve(),
        }
    )
    for __path in sys_path_entries:
        for __valid_path_prefix in __valid_path_prefixes:
            try:
                __resolved_path = pathlib.Path(__path).resolve()
                __resolved_path.relative_to(__valid_path_prefix)
                __sys_path.append(str(__resolved_path))
            except ValueError:
                continue
    if "PYTHONPATH" in os.environ:
        for __path in os.environ["PYTHONPATH"].split(os.pathsep):
            __sys_path.append(__path)
    for known_path in __sys_path[:]:
        for _ in pathlib.Path(known_path).glob("__editable__.*.pth"):
            paths = addpackage(known_path, _)
            for p in paths:
                if p not in __sys_path:
                    __sys_path.append(p)
    return __sys_path



"""
    salt.serializers
    ~~~~~~~~~~~~~~~~~~~~~~

    This module implements all the serializers needed by salt.
    Each serializer offers the same functions and attributes:

    :deserialize: function for deserializing string or stream

    :serialize: function for serializing a Python object

    :available: flag that tells if the serializer is available
                (all dependencies are met etc.)

"""

from salt.exceptions import SaltException, SaltRenderError


class DeserializationError(SaltRenderError, RuntimeError):
    """Raised when stream of string failed to be deserialized"""


class SerializationError(SaltException, RuntimeError):
    """Raised when stream of string failed to be serialized"""



"""
A state module to manage Cisco UCS chassis devices.

:codeauthor: ``Spencer Ervin <spencer_ervin@hotmail.com>``
:maturity:   new
:depends:    none
:platform:   unix


About
=====
This state module was designed to handle connections to a Cisco Unified Computing System (UCS) chassis. This module
relies on the CIMC proxy module to interface with the device.

.. seealso::
    :py:mod:`CIMC Proxy Module <salt.proxy.cimc>`

"""

import logging

log = logging.getLogger(__name__)


def __virtual__():
    if "cimc.get_system_info" in __salt__:
        return True
    return (False, "cimc module could not be loaded")


def _default_ret(name):
    """
    Set the default response values.

    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}
    return ret


def hostname(name, hostname=None):
    """
    Ensures that the hostname is set to the specified value.

    .. versionadded:: 2019.2.0

    name: The name of the module function to execute.

    hostname(str): The hostname of the server.

    SLS Example:

    .. code-block:: yaml

        set_name:
          cimc.hostname:
            - hostname: foobar

    """

    ret = _default_ret(name)

    current_name = __salt__["cimc.get_hostname"]()

    req_change = False

    try:

        if current_name != hostname:
            req_change = True

        if req_change:

            update = __salt__["cimc.set_hostname"](hostname)

            if not update:
                ret["result"] = False
                ret["comment"] = "Error setting hostname."
                return ret

            ret["changes"]["before"] = current_name
            ret["changes"]["after"] = hostname
            ret["comment"] = "Hostname modified."
        else:
            ret["comment"] = "Hostname already configured. No changes required."

    except Exception as err:  # pylint: disable=broad-except
        ret["result"] = False
        ret["comment"] = "Error occurred setting hostname."
        log.error(err)
        return ret

    ret["result"] = True

    return ret


def logging_levels(name, remote=None, local=None):
    """
    Ensures that the logging levels are set on the device. The logging levels
    must match the following options: emergency, alert, critical, error, warning,
    notice, informational, debug.

    .. versionadded:: 2019.2.0

    name: The name of the module function to execute.

    remote(str): The logging level for SYSLOG logs.

    local(str): The logging level for the local device.

    SLS Example:

    .. code-block:: yaml

        logging_levels:
          cimc.logging_levels:
            - remote: informational
            - local: notice

    """

    ret = _default_ret(name)

    syslog_conf = __salt__["cimc.get_syslog_settings"]()

    req_change = False

    try:
        syslog_dict = syslog_conf["outConfigs"]["commSyslog"][0]

        if remote and syslog_dict["remoteSeverity"] != remote:
            req_change = True
        elif local and syslog_dict["localSeverity"] != local:
            req_change = True

        if req_change:

            update = __salt__["cimc.set_logging_levels"](remote, local)

            if update["outConfig"]["commSyslog"][0]["status"] != "modified":
                ret["result"] = False
                ret["comment"] = "Error setting logging levels."
                return ret

            ret["changes"]["before"] = syslog_conf
            ret["changes"]["after"] = __salt__["cimc.get_syslog_settings"]()
            ret["comment"] = "Logging level settings modified."
        else:
            ret["comment"] = "Logging level already configured. No changes required."

    except Exception as err:  # pylint: disable=broad-except
        ret["result"] = False
        ret["comment"] = "Error occurred setting logging level settings."
        log.error(err)
        return ret

    ret["result"] = True

    return ret


def ntp(name, servers):
    """
    Ensures that the NTP servers are configured. Servers are provided as an individual string or list format. Only four
    NTP servers will be reviewed. Any entries past four will be ignored.

    name: The name of the module function to execute.

    servers(str, list): The IP address or FQDN of the NTP servers.

    SLS Example:

    .. code-block:: yaml

        ntp_configuration_list:
          cimc.ntp:
            - servers:
              - foo.bar.com
              - 10.10.10.10

        ntp_configuration_str:
          cimc.ntp:
            - servers: foo.bar.com

    """
    ret = _default_ret(name)

    ntp_servers = ["", "", "", ""]

    # Parse our server arguments
    if isinstance(servers, list):
        i = 0
        for x in servers:
            ntp_servers[i] = x
            i += 1
    else:
        ntp_servers[0] = servers

    conf = __salt__["cimc.get_ntp"]()

    # Check if our NTP configuration is already set
    req_change = False
    try:
        if (
            conf["outConfigs"]["commNtpProvider"][0]["ntpEnable"] != "yes"
            or ntp_servers[0] != conf["outConfigs"]["commNtpProvider"][0]["ntpServer1"]
            or ntp_servers[1] != conf["outConfigs"]["commNtpProvider"][0]["ntpServer2"]
            or ntp_servers[2] != conf["outConfigs"]["commNtpProvider"][0]["ntpServer3"]
            or ntp_servers[3] != conf["outConfigs"]["commNtpProvider"][0]["ntpServer4"]
        ):
            req_change = True
    except KeyError as err:
        ret["result"] = False
        ret["comment"] = "Unable to confirm current NTP settings."
        log.error(err)
        return ret

    if req_change:

        try:
            update = __salt__["cimc.set_ntp_server"](
                ntp_servers[0], ntp_servers[1], ntp_servers[2], ntp_servers[3]
            )
            if update["outConfig"]["commNtpProvider"][0]["status"] != "modified":
                ret["result"] = False
                ret["comment"] = "Error setting NTP configuration."
                return ret
        except Exception as err:  # pylint: disable=broad-except
            ret["result"] = False
            ret["comment"] = "Error setting NTP configuration."
            log.error(err)
            return ret

        ret["changes"]["before"] = conf
        ret["changes"]["after"] = __salt__["cimc.get_ntp"]()
        ret["comment"] = "NTP settings modified."
    else:
        ret["comment"] = "NTP already configured. No changes required."

    ret["result"] = True

    return ret


def power_configuration(name, policy=None, delayType=None, delayValue=None):
    """
    Ensures that the power configuration is configured on the system. This is
    only available on some C-Series servers.

    .. versionadded:: 2019.2.0

    name: The name of the module function to execute.

    policy(str): The action to be taken when chassis power is restored after
    an unexpected power loss. This can be one of the following:

        reset: The server is allowed to boot up normally when power is
        restored. The server can restart immediately or, optionally, after a
        fixed or random delay.

        stay-off: The server remains off until it is manually restarted.

        last-state: The server restarts and the system attempts to restore
        any processes that were running before power was lost.

    delayType(str): If the selected policy is reset, the restart can be
    delayed with this option. This can be one of the following:

        fixed: The server restarts after a fixed delay.

        random: The server restarts after a random delay.

    delayValue(int): If a fixed delay is selected, once chassis power is
    restored and the Cisco IMC has finished rebooting, the system waits for
    the specified number of seconds before restarting the server. Enter an
    integer between 0 and 240.


    SLS Example:

    .. code-block:: yaml

        reset_power:
          cimc.power_configuration:
            - policy: reset
            - delayType: fixed
            - delayValue: 0

        power_off:
          cimc.power_configuration:
            - policy: stay-off


    """

    ret = _default_ret(name)

    power_conf = __salt__["cimc.get_power_configuration"]()

    req_change = False

    try:
        power_dict = power_conf["outConfigs"]["biosVfResumeOnACPowerLoss"][0]

        if policy and power_dict["vpResumeOnACPowerLoss"] != policy:
            req_change = True
        elif policy == "reset":
            if power_dict["delayType"] != delayType:
                req_change = True
            elif power_dict["delayType"] == "fixed":
                if str(power_dict["delay"]) != str(delayValue):
                    req_change = True
        else:
            ret["result"] = False
            ret["comment"] = "The power policy must be specified."
            return ret

        if req_change:

            update = __salt__["cimc.set_power_configuration"](
                policy, delayType, delayValue
            )

            if (
                update["outConfig"]["biosVfResumeOnACPowerLoss"][0]["status"]
                != "modified"
            ):
                ret["result"] = False
                ret["comment"] = "Error setting power configuration."
                return ret

            ret["changes"]["before"] = power_conf
            ret["changes"]["after"] = __salt__["cimc.get_power_configuration"]()
            ret["comment"] = "Power settings modified."
        else:
            ret["comment"] = "Power settings already configured. No changes required."

    except Exception as err:  # pylint: disable=broad-except
        ret["result"] = False
        ret["comment"] = "Error occurred setting power settings."
        log.error(err)
        return ret

    ret["result"] = True

    return ret


def syslog(name, primary=None, secondary=None):
    """
    Ensures that the syslog servers are set to the specified values. A value of None will be ignored.

    name: The name of the module function to execute.

    primary(str): The IP address or FQDN of the primary syslog server.

    secondary(str): The IP address or FQDN of the secondary syslog server.

    SLS Example:

    .. code-block:: yaml

        syslog_configuration:
          cimc.syslog:
            - primary: 10.10.10.10
            - secondary: foo.bar.com

    """
    ret = _default_ret(name)

    conf = __salt__["cimc.get_syslog"]()

    req_change = False

    if primary:
        prim_change = True
        if "outConfigs" in conf and "commSyslogClient" in conf["outConfigs"]:
            for entry in conf["outConfigs"]["commSyslogClient"]:
                if entry["name"] != "primary":
                    continue
                if entry["adminState"] == "enabled" and entry["hostname"] == primary:
                    prim_change = False

        if prim_change:
            try:
                update = __salt__["cimc.set_syslog_server"](primary, "primary")
                if update["outConfig"]["commSyslogClient"][0]["status"] == "modified":
                    req_change = True
                else:
                    ret["result"] = False
                    ret["comment"] = "Error setting primary SYSLOG server."
                    return ret
            except Exception as err:  # pylint: disable=broad-except
                ret["result"] = False
                ret["comment"] = "Error setting primary SYSLOG server."
                log.error(err)
                return ret

    if secondary:
        sec_change = True
        if "outConfig" in conf and "commSyslogClient" in conf["outConfig"]:
            for entry in conf["outConfig"]["commSyslogClient"]:
                if entry["name"] != "secondary":
                    continue
                if entry["adminState"] == "enabled" and entry["hostname"] == secondary:
                    sec_change = False

        if sec_change:
            try:
                update = __salt__["cimc.set_syslog_server"](secondary, "secondary")
                if update["outConfig"]["commSyslogClient"][0]["status"] == "modified":
                    req_change = True
                else:
                    ret["result"] = False
                    ret["comment"] = "Error setting secondary SYSLOG server."
                    return ret
            except Exception as err:  # pylint: disable=broad-except
                ret["result"] = False
                ret["comment"] = "Error setting secondary SYSLOG server."
                log.error(err)
                return ret

    if req_change:
        ret["changes"]["before"] = conf
        ret["changes"]["after"] = __salt__["cimc.get_syslog"]()
        ret["comment"] = "SYSLOG settings modified."
    else:
        ret["comment"] = "SYSLOG already configured. No changes required."

    ret["result"] = True

    return ret


def user(name, id="", user="", priv="", password="", status="active"):
    """
    Ensures that a user is configured on the device. Due to being unable to
    verify the user password. This is a forced operation.

    .. versionadded:: 2019.2.0

    name: The name of the module function to execute.

    id(int): The user ID slot on the device.

    user(str): The username of the user.

    priv(str): The privilege level of the user.

    password(str): The password of the user.

    status(str): The status of the user. Can be either active or inactive.

    SLS Example:

    .. code-block:: yaml

        user_configuration:
          cimc.user:
            - id: 11
            - user: foo
            - priv: admin
            - password: mypassword
            - status: active

    """

    ret = _default_ret(name)

    user_conf = __salt__["cimc.get_users"]()

    try:
        for entry in user_conf["outConfigs"]["aaaUser"]:
            if entry["id"] == str(id):
                conf = entry

        if not conf:
            ret["result"] = False
            ret["comment"] = (
                "Unable to find requested user id on device. Please verify id is valid."
            )
            return ret

        updates = __salt__["cimc.set_user"](str(id), user, password, priv, status)

        if "outConfig" in updates:
            ret["changes"]["before"] = conf
            ret["changes"]["after"] = updates["outConfig"]["aaaUser"]
            ret["comment"] = "User settings modified."
        else:
            ret["result"] = False
            ret["comment"] = "Error setting user configuration."
            return ret

    except Exception as err:  # pylint: disable=broad-except
        ret["result"] = False
        ret["comment"] = "Error setting user configuration."
        log.error(err)
        return ret

    ret["result"] = True

    return ret



"""
Manage Grafana v4.0 orgs

.. versionadded:: 2017.7.0

:configuration: This state requires a configuration profile to be configured
    in the minion config, minion pillar, or master config. The module will use
    the 'grafana' key by default, if defined.

    Example configuration using basic authentication:

    .. code-block:: yaml

        grafana:
          grafana_url: http://grafana.localhost
          grafana_user: admin
          grafana_password: admin
          grafana_timeout: 3

    Example configuration using token based authentication:

    .. code-block:: yaml

        grafana:
          grafana_url: http://grafana.localhost
          grafana_token: token
          grafana_timeout: 3

.. code-block:: yaml

    Ensure foobar org is present:
      grafana4_org.present:
        - name: foobar
        - theme:  ""
        - home_dashboard_id: 0
        - timezone: "utc"
        - address1: ""
        - address2: ""
        - city: ""
        - zip_code: ""
        - state: ""
        - country: ""
"""

from requests.exceptions import HTTPError

import salt.utils.dictupdate as dictupdate
from salt.utils.dictdiffer import deep_diff


def __virtual__():
    """Only load if grafana4 module is available"""
    if "grafana4.get_org" in __salt__:
        return True
    return (False, "grafana4 module could not be loaded")


def present(
    name,
    users=None,
    theme=None,
    home_dashboard_id=None,
    timezone=None,
    address1=None,
    address2=None,
    city=None,
    zip_code=None,
    address_state=None,
    country=None,
    profile="grafana",
):
    """
    Ensure that an organization is present.

    name
        Name of the org.

    users
        Optional - Dict of user/role associated with the org. Example:

        .. code-block:: yaml

            users:
              foo: Viewer
              bar: Editor

    theme
        Optional - Selected theme for the org.

    home_dashboard_id
        Optional - Home dashboard for the org.

    timezone
        Optional - Timezone for the org (one of: "browser", "utc", or "").

    address1
        Optional - address1 of the org.

    address2
        Optional - address2 of the org.

    city
        Optional - city of the org.

    zip_code
        Optional - zip_code of the org.

    address_state
        Optional - state of the org.

    country
        Optional - country of the org.

    profile
        Configuration profile used to connect to the Grafana instance.
        Default is 'grafana'.
    """
    if isinstance(profile, str):
        profile = __salt__["config.option"](profile)

    ret = {"name": name, "result": None, "comment": None, "changes": {}}
    create = False
    try:
        org = __salt__["grafana4.get_org"](name, profile)
    except HTTPError as e:
        if e.response.status_code == 404:
            create = True
        else:
            raise

    if create:
        if __opts__["test"]:
            ret["comment"] = f"Org {name} will be created"
            return ret
        __salt__["grafana4.create_org"](profile=profile, name=name)
        org = __salt__["grafana4.get_org"](name, profile)
        ret["changes"] = org
        ret["comment"] = f"New org {name} added"

    data = _get_json_data(
        address1=address1,
        address2=address2,
        city=city,
        zipCode=zip_code,
        state=address_state,
        country=country,
        defaults=org["address"],
    )
    if data != org["address"]:
        if __opts__["test"]:
            ret["comment"] = f"Org {name} address will be updated"
            return ret
        __salt__["grafana4.update_org_address"](name, profile=profile, **data)
        if create:
            dictupdate.update(ret["changes"]["address"], data)
        else:
            dictupdate.update(ret["changes"], deep_diff(org["address"], data))

    prefs = __salt__["grafana4.get_org_prefs"](name, profile=profile)
    data = _get_json_data(
        theme=theme,
        homeDashboardId=home_dashboard_id,
        timezone=timezone,
        defaults=prefs,
    )
    if data != prefs:
        if __opts__["test"]:
            ret["comment"] = f"Org {name} prefs will be updated"
            return ret
        __salt__["grafana4.update_org_prefs"](name, profile=profile, **data)
        if create:
            dictupdate.update(ret["changes"], data)
        else:
            dictupdate.update(ret["changes"], deep_diff(prefs, data))

    if users:
        db_users = {}
        for item in __salt__["grafana4.get_org_users"](name, profile=profile):
            db_users[item["login"]] = {
                "userId": item["userId"],
                "role": item["role"],
            }
        for username, role in users.items():
            if username in db_users:
                if role is False:
                    if __opts__["test"]:
                        ret["comment"] = "Org {} user {} will be deleted".format(
                            name, username
                        )
                        return ret
                    __salt__["grafana4.delete_org_user"](
                        db_users[username]["userId"], profile=profile
                    )
                elif role != db_users[username]["role"]:
                    if __opts__["test"]:
                        ret["comment"] = "Org {} user {} role will be updated".format(
                            name, username
                        )
                        return ret
                    __salt__["grafana4.update_org_user"](
                        db_users[username]["userId"],
                        loginOrEmail=username,
                        role=role,
                        profile=profile,
                    )
            elif role:
                if __opts__["test"]:
                    ret["comment"] = "Org {} user {} will be created".format(
                        name, username
                    )
                    return ret
                __salt__["grafana4.create_org_user"](
                    loginOrEmail=username, role=role, profile=profile
                )

        new_db_users = {}
        for item in __salt__["grafana4.get_org_users"](name, profile=profile):
            new_db_users[item["login"]] = {
                "userId": item["userId"],
                "role": item["role"],
            }
        if create:
            dictupdate.update(ret["changes"], new_db_users)
        else:
            dictupdate.update(ret["changes"], deep_diff(db_users, new_db_users))

    ret["result"] = True
    if not create:
        if ret["changes"]:
            ret["comment"] = f"Org {name} updated"
        else:
            ret["changes"] = {}
            ret["comment"] = f"Org {name} already up-to-date"

    return ret


def absent(name, profile="grafana"):
    """
    Ensure that a org is present.

    name
        Name of the org to remove.

    profile
        Configuration profile used to connect to the Grafana instance.
        Default is 'grafana'.
    """
    if isinstance(profile, str):
        profile = __salt__["config.option"](profile)

    ret = {"name": name, "result": None, "comment": None, "changes": {}}
    org = __salt__["grafana4.get_org"](name, profile)

    if not org:
        ret["result"] = True
        ret["comment"] = f"Org {name} already absent"
        return ret

    if __opts__["test"]:
        ret["comment"] = f"Org {name} will be deleted"
        return ret
    __salt__["grafana4.delete_org"](org["id"], profile=profile)

    ret["result"] = True
    ret["changes"][name] = "Absent"
    ret["comment"] = f"Org {name} was deleted"

    return ret


def _get_json_data(defaults=None, **kwargs):
    if defaults is None:
        defaults = {}
    for k, v in kwargs.items():
        if v is None:
            kwargs[k] = defaults.get(k)
    return kwargs



"""
Salt-specific interface for calling Salt Cloud directly
"""

import copy
import logging
import os

import salt.utils.data
from salt.exceptions import SaltCloudConfigError

try:
    import salt.cloud

    HAS_SALTCLOUD = True
except ImportError:
    HAS_SALTCLOUD = False


log = logging.getLogger(__name__)

__func_alias__ = {"profile_": "profile"}


def __virtual__():
    """
    Only work on POSIX-like systems
    """
    if HAS_SALTCLOUD:
        return True
    return (
        False,
        "The cloud execution module cannot be loaded: only available on non-Windows"
        " systems.",
    )


def _get_client():
    """
    Return a cloud client
    """
    client = salt.cloud.CloudClient(
        os.path.join(os.path.dirname(__opts__["conf_file"]), "cloud"),
        pillars=copy.deepcopy(__pillar__.get("cloud", {})),
    )
    return client


def list_sizes(provider="all"):
    """
    List cloud provider sizes for the given providers

    CLI Example:

    .. code-block:: bash

        salt minionname cloud.list_sizes my-gce-config
    """
    client = _get_client()
    sizes = client.list_sizes(provider)
    return sizes


def list_images(provider="all"):
    """
    List cloud provider images for the given providers

    CLI Example:

    .. code-block:: bash

        salt minionname cloud.list_images my-gce-config
    """
    client = _get_client()
    images = client.list_images(provider)
    return images


def list_locations(provider="all"):
    """
    List cloud provider locations for the given providers

    CLI Example:

    .. code-block:: bash

        salt minionname cloud.list_locations my-gce-config
    """
    client = _get_client()
    locations = client.list_locations(provider)
    return locations


def query(query_type="list_nodes"):
    """
    List cloud provider data for all providers

    CLI Examples:

    .. code-block:: bash

        salt minionname cloud.query
        salt minionname cloud.query list_nodes_full
        salt minionname cloud.query list_nodes_select
    """
    client = _get_client()
    info = client.query(query_type)
    return info


def full_query(query_type="list_nodes_full"):
    """
    List all available cloud provider data

    CLI Example:

    .. code-block:: bash

        salt minionname cloud.full_query
    """
    return query(query_type=query_type)


def select_query(query_type="list_nodes_select"):
    """
    List selected nodes

    CLI Example:

    .. code-block:: bash

        salt minionname cloud.select_query
    """
    return query(query_type=query_type)


def has_instance(name, provider=None):
    """
    Return true if the instance is found on a provider

    CLI Example:

    .. code-block:: bash

        salt minionname cloud.has_instance myinstance
    """
    data = get_instance(name, provider)
    if data is None:
        return False
    return True


def get_instance(name, provider=None):
    """
    Return details on an instance.

    Similar to the cloud action show_instance
    but returns only the instance details.

    CLI Example:

    .. code-block:: bash

        salt minionname cloud.get_instance myinstance

    SLS Example:

    .. code-block:: bash

        {{ salt['cloud.get_instance']('myinstance')['mac_address'] }}

    """
    data = action(fun="show_instance", names=[name], provider=provider)
    info = salt.utils.data.simple_types_filter(data)
    try:
        # get the first: [alias][driver][vm_name]
        info = next(iter(next(iter(next(iter(info.values())).values())).values()))
    except AttributeError:
        return None
    return info


def profile_(profile, names, vm_overrides=None, opts=None, **kwargs):
    """
    Spin up an instance using Salt Cloud

    CLI Example:

    .. code-block:: bash

        salt minionname cloud.profile my-gce-config myinstance
    """
    client = _get_client()
    if isinstance(opts, dict):
        client.opts.update(opts)
    info = client.profile(profile, names, vm_overrides=vm_overrides, **kwargs)
    return info


def map_run(path=None, **kwargs):
    """
    Execute a salt cloud map file

    Cloud Map data can be retrieved from several sources:

    - a local file (provide the path to the file to the 'path' argument)
    - a JSON-formatted map directly (provide the appropriately formatted to using the 'map_data' argument)
    - the Salt Pillar (provide the map name of under 'pillar:cloud:maps' to the 'map_pillar' argument)

    .. note::
        Only one of these sources can be read at a time. The options are listed
        in their order of precedence.

    CLI Examples:

    .. code-block:: bash

        salt minionname cloud.map_run /path/to/cloud.map
        salt minionname cloud.map_run path=/path/to/cloud.map
        salt minionname cloud.map_run map_pillar='<map_pillar>'
          .. versionchanged:: 2018.3.1
        salt minionname cloud.map_run map_data='<actual map data>'
    """
    client = _get_client()
    info = client.map_run(path, **kwargs)
    return info


def destroy(names):
    """
    Destroy the named VM(s)

    CLI Example:

    .. code-block:: bash

        salt minionname cloud.destroy myinstance
    """
    client = _get_client()
    info = client.destroy(names)
    return info


def action(fun=None, cloudmap=None, names=None, provider=None, instance=None, **kwargs):
    """
    Execute a single action on the given provider/instance

    CLI Example:

    .. code-block:: bash

        salt minionname cloud.action start instance=myinstance
        salt minionname cloud.action stop instance=myinstance
        salt minionname cloud.action show_image provider=my-ec2-config image=ami-1624987f
    """
    client = _get_client()
    try:
        info = client.action(fun, cloudmap, names, provider, instance, kwargs)
    except SaltCloudConfigError as err:
        log.error(err)
        return None

    return info


def create(provider, names, opts=None, **kwargs):
    """
    Create an instance using Salt Cloud

    CLI Example:

    .. code-block:: bash

        salt minionname cloud.create my-ec2-config myinstance image=ami-1624987f size='t1.micro' ssh_username=ec2-user securitygroup=default delvol_on_destroy=True
    """
    client = _get_client()
    if isinstance(opts, dict):
        client.opts.update(opts)
    info = client.create(provider, names, **kwargs)
    return info


def volume_list(provider):
    """
    List block storage volumes

    CLI Example:

    .. code-block:: bash

        salt minionname cloud.volume_list my-nova

    """
    client = _get_client()
    info = client.extra_action(action="volume_list", provider=provider, names="name")
    return info["name"]


def volume_delete(provider, names, **kwargs):
    """
    Delete volume

    CLI Example:

    .. code-block:: bash

        salt minionname cloud.volume_delete my-nova myblock

    """
    client = _get_client()
    info = client.extra_action(
        provider=provider, names=names, action="volume_delete", **kwargs
    )
    return info


def volume_create(provider, names, **kwargs):
    """
    Create volume

    CLI Example:

    .. code-block:: bash

        salt minionname cloud.volume_create my-nova myblock size=100 voltype=SSD

    """
    client = _get_client()
    info = client.extra_action(
        action="volume_create", names=names, provider=provider, **kwargs
    )
    return info


def volume_attach(provider, names, **kwargs):
    """
    Attach volume to a server

    CLI Example:

    .. code-block:: bash

        salt minionname cloud.volume_attach my-nova myblock server_name=myserver device='/dev/xvdf'

    """
    client = _get_client()
    info = client.extra_action(
        provider=provider, names=names, action="volume_attach", **kwargs
    )
    return info


def volume_detach(provider, names, **kwargs):
    """
    Detach volume from a server

    CLI Example:

    .. code-block:: bash

        salt minionname cloud.volume_detach my-nova myblock server_name=myserver

    """
    client = _get_client()
    info = client.extra_action(
        provider=provider, names=names, action="volume_detach", **kwargs
    )
    return info


def network_list(provider):
    """
    List private networks

    CLI Example:

    .. code-block:: bash

        salt minionname cloud.network_list my-nova

    """
    client = _get_client()
    return client.extra_action(action="network_list", provider=provider, names="names")


def network_create(provider, names, **kwargs):
    """
    Create private network

    CLI Example:

    .. code-block:: bash

        salt minionname cloud.network_create my-nova names=['salt'] cidr='192.168.100.0/24'

    """
    client = _get_client()
    return client.extra_action(
        provider=provider, names=names, action="network_create", **kwargs
    )


def virtual_interface_list(provider, names, **kwargs):
    """
    List virtual interfaces on a server

    CLI Example:

    .. code-block:: bash

        salt minionname cloud.virtual_interface_list my-nova names=['salt-master']

    """
    client = _get_client()
    return client.extra_action(
        provider=provider, names=names, action="virtual_interface_list", **kwargs
    )


def virtual_interface_create(provider, names, **kwargs):
    """
    Attach private interfaces to a server

    CLI Example:

    .. code-block:: bash

        salt minionname cloud.virtual_interface_create my-nova names=['salt-master'] net_name='salt'

    """
    client = _get_client()
    return client.extra_action(
        provider=provider, names=names, action="virtual_interface_create", **kwargs
    )



"""
ZFS grain provider

:maintainer:    Jorge Schrauwen <sjorge@blackdot.be>
:maturity:      new
:depends:       salt.module.cmdmod
:platform:      illumos,freebsd,linux

.. versionadded:: 2018.3.0

"""

import logging

# Solve the Chicken and egg problem where grains need to run before any
# of the modules are loaded and are generally available for any usage.
import salt.modules.cmdmod
import salt.utils.dictupdate
import salt.utils.path
import salt.utils.platform
import salt.utils.zfs

__virtualname__ = "zfs"
__salt__ = {
    "cmd.run": salt.modules.cmdmod.run,
}
__utils__ = {
    "zfs.is_supported": salt.utils.zfs.is_supported,
    "zfs.has_feature_flags": salt.utils.zfs.has_feature_flags,
    "zfs.zpool_command": salt.utils.zfs.zpool_command,
    "zfs.to_size": salt.utils.zfs.to_size,
}

log = logging.getLogger(__name__)


def __virtual__():
    """
    Load zfs grains
    """
    # NOTE: we always load this grain so we can properly export
    #       at least the zfs_support grain
    #       except for Windows... don't try to load this on Windows (#51703)
    if salt.utils.platform.is_windows():
        return False, "ZFS: Not available on Windows"
    return __virtualname__


def _zfs_pool_data():
    """
    Provide grains about zpools
    """
    grains = {}

    # collect zpool data
    zpool_list_cmd = __utils__["zfs.zpool_command"](
        "list",
        flags=["-H"],
        opts={"-o": "name,size"},
    )
    for zpool in __salt__["cmd.run"](zpool_list_cmd, ignore_retcode=True).splitlines():
        if "zpool" not in grains:
            grains["zpool"] = {}
        zpool = zpool.split()
        grains["zpool"][zpool[0]] = __utils__["zfs.to_size"](zpool[1], False)

    # return grain data
    return grains


def zfs():
    """
    Provide grains for zfs/zpool
    """
    grains = {}
    grains["zfs_support"] = __utils__["zfs.is_supported"]()
    grains["zfs_feature_flags"] = __utils__["zfs.has_feature_flags"]()
    if grains["zfs_support"]:
        grains = salt.utils.dictupdate.update(
            grains, _zfs_pool_data(), merge_lists=True
        )

    return grains



"""
Control Linux Containers via Salt

:depends: lxc execution module
"""

import copy
import logging
import os
import time

import salt.client
import salt.key
import salt.utils.args
import salt.utils.cloud
import salt.utils.files
import salt.utils.stringutils
import salt.utils.virt
from salt.utils.odict import OrderedDict as _OrderedDict

log = logging.getLogger(__name__)

# Don't shadow built-in's.
__func_alias__ = {"list_": "list"}


def _do(name, fun, path=None):
    """
    Invoke a function in the lxc module with no args

    path
        path to the container parent
        default: /var/lib/lxc (system default)

        .. versionadded:: 2015.8.0
    """
    host = find_guest(name, quiet=True, path=path)
    if not host:
        return False

    with salt.client.get_local_client(__opts__["conf_file"]) as client:
        cmd_ret = client.cmd_iter(
            host, f"lxc.{fun}", [name], kwarg={"path": path}, timeout=60
        )
        data = next(cmd_ret)
        data = data.get(host, {}).get("ret", None)
        if data:
            data = {host: data}
        return data


def _do_names(names, fun, path=None):
    """
    Invoke a function in the lxc module with no args

    path
        path to the container parent
        default: /var/lib/lxc (system default)

        .. versionadded:: 2015.8.0
    """
    ret = {}
    hosts = find_guests(names, path=path)
    if not hosts:
        return False

    with salt.client.get_local_client(__opts__["conf_file"]) as client:
        for host, sub_names in hosts.items():
            cmds = []
            for name in sub_names:
                cmds.append(
                    client.cmd_iter(
                        host,
                        f"lxc.{fun}",
                        [name],
                        kwarg={"path": path},
                        timeout=60,
                    )
                )
            for cmd in cmds:
                data = next(cmd)
                data = data.get(host, {}).get("ret", None)
                if data:
                    ret.update({host: data})
        return ret


def find_guest(name, quiet=False, path=None):
    """
    Returns the host for a container.

    path
        path to the container parent
        default: /var/lib/lxc (system default)

        .. versionadded:: 2015.8.0


    .. code-block:: bash

        salt-run lxc.find_guest name
    """
    if quiet:
        log.warning("'quiet' argument is being deprecated. Please migrate to --quiet")
    for data in _list_iter(path=path):
        host, l = next(iter(data.items()))
        for x in "running", "frozen", "stopped":
            if name in l[x]:
                if not quiet:
                    __jid_event__.fire_event(
                        {"data": host, "outputter": "lxc_find_host"}, "progress"
                    )
                return host
    return None


def find_guests(names, path=None):
    """
    Return a dict of hosts and named guests

    path
        path to the container parent
        default: /var/lib/lxc (system default)

        .. versionadded:: 2015.8.0

    """
    ret = {}
    names = names.split(",")
    for data in _list_iter(path=path):
        host, stat = next(iter(data.items()))
        for state in stat:
            for name in stat[state]:
                if name in names:
                    if host in ret:
                        ret[host].append(name)
                    else:
                        ret[host] = [name]
    return ret


def init(names, host=None, saltcloud_mode=False, quiet=False, **kwargs):
    """
    Initialize a new container


    .. code-block:: bash

        salt-run lxc.init name host=minion_id [cpuset=cgroups_cpuset] \\
                [cpushare=cgroups_cpushare] [memory=cgroups_memory] \\
                [template=lxc_template_name] [clone=original name] \\
                [profile=lxc_profile] [network_proflile=network_profile] \\
                [nic=network_profile] [nic_opts=nic_opts] \\
                [start=(true|false)] [seed=(true|false)] \\
                [install=(true|false)] [config=minion_config] \\
                [snapshot=(true|false)]

    names
        Name of the containers, supports a single name or a comma delimited
        list of names.

    host
        Minion on which to initialize the container **(required)**

    path
        path to the container parent
        default: /var/lib/lxc (system default)

        .. versionadded:: 2015.8.0

    saltcloud_mode
        init the container with the saltcloud opts format instead
        See lxc.init_interface module documentation

    cpuset
        cgroups cpuset.

    cpushare
        cgroups cpu shares.

    memory
        cgroups memory limit, in MB

        .. versionchanged:: 2015.5.0
            If no value is passed, no limit is set. In earlier Salt versions,
            not passing this value causes a 1024MB memory limit to be set, and
            it was necessary to pass ``memory=0`` to set no limit.

    template
        Name of LXC template on which to base this container

    clone
        Clone this container from an existing container

    profile
        A LXC profile (defined in config or pillar).

    network_profile
        Network profile to use for the container

        .. versionadded:: 2015.5.2

    nic
        .. deprecated:: 2015.5.0
            Use ``network_profile`` instead

    nic_opts
        Extra options for network interfaces. E.g.:

        ``{"eth0": {"mac": "aa:bb:cc:dd:ee:ff", "ipv4": "10.1.1.1", "ipv6": "2001:db8::ff00:42:8329"}}``

    start
        Start the newly created container.

    seed
        Seed the container with the minion config and autosign its key.
        Default: true

    install
        If salt-minion is not already installed, install it. Default: true

    config
        Optional config parameters. By default, the id is set to
        the name of the container.
    """
    path = kwargs.get("path", None)
    if quiet:
        log.warning("'quiet' argument is being deprecated. Please migrate to --quiet")
    ret = {"comment": "", "result": True}
    if host is None:
        # TODO: Support selection of host based on available memory/cpu/etc.
        ret["comment"] = "A host must be provided"
        ret["result"] = False
        return ret
    if isinstance(names, str):
        names = names.split(",")
    if not isinstance(names, list):
        ret["comment"] = "Container names are not formed as a list"
        ret["result"] = False
        return ret
    # check that the host is alive
    alive = False
    with salt.client.get_local_client(__opts__["conf_file"]) as client:
        try:
            if client.cmd(host, "test.ping", timeout=20).get(host, None):
                alive = True
        except (TypeError, KeyError):
            pass
        if not alive:
            ret["comment"] = f"Host {host} is not reachable"
            ret["result"] = False
            return ret

        log.info("Searching for LXC Hosts")
        data = __salt__["lxc.list"](host, quiet=True, path=path)
        for host, containers in data.items():
            for name in names:
                if name in sum(containers.values(), []):
                    log.info(
                        "Container '%s' already exists on host '%s', init "
                        "can be a NO-OP",
                        name,
                        host,
                    )
        if host not in data:
            ret["comment"] = f"Host '{host}' was not found"
            ret["result"] = False
            return ret

        kw = salt.utils.args.clean_kwargs(**kwargs)
        pub_key = kw.get("pub_key", None)
        priv_key = kw.get("priv_key", None)
        explicit_auth = pub_key and priv_key
        approve_key = kw.get("approve_key", True)
        seeds = {}
        seed_arg = kwargs.get("seed", True)
        if approve_key and not explicit_auth:
            with salt.key.Key(__opts__) as skey:
                all_minions = skey.all_keys().get("minions", [])
                for name in names:
                    seed = seed_arg
                    if name in all_minions:
                        try:
                            if client.cmd(name, "test.ping", timeout=20).get(
                                name, None
                            ):
                                seed = False
                        except (TypeError, KeyError):
                            pass
                    seeds[name] = seed
                    kv = salt.utils.virt.VirtKey(host, name, __opts__)
                    if kv.authorize():
                        log.info("Container key will be preauthorized")
                    else:
                        ret["comment"] = "Container key preauthorization failed"
                        ret["result"] = False
                        return ret

        log.info("Creating container(s) '%s' on host '%s'", names, host)

        cmds = []
        for name in names:
            args = [name]
            kw = salt.utils.args.clean_kwargs(**kwargs)
            if saltcloud_mode:
                kw = copy.deepcopy(kw)
                kw["name"] = name
                saved_kwargs = kw
                kw = client.cmd(
                    host,
                    "lxc.cloud_init_interface",
                    args + [kw],
                    tgt_type="list",
                    timeout=600,
                ).get(host, {})
                kw.update(saved_kwargs)
            name = kw.pop("name", name)
            # be sure not to seed an already seeded host
            kw["seed"] = seeds.get(name, seed_arg)
            if not kw["seed"]:
                kw.pop("seed_cmd", "")
            cmds.append(
                (
                    host,
                    name,
                    client.cmd_iter(host, "lxc.init", args, kwarg=kw, timeout=600),
                )
            )
        done = ret.setdefault("done", [])
        errors = ret.setdefault("errors", _OrderedDict())

        for ix, acmd in enumerate(cmds):
            hst, container_name, cmd = acmd
            containers = ret.setdefault(hst, [])
            herrs = errors.setdefault(hst, _OrderedDict())
            serrs = herrs.setdefault(container_name, [])
            sub_ret = next(cmd)
            error = None
            if isinstance(sub_ret, dict) and host in sub_ret:
                j_ret = sub_ret[hst]
                container = j_ret.get("ret", {})
                if container and isinstance(container, dict):
                    if not container.get("result", False):
                        error = container
                else:
                    error = "Invalid return for {}: {} {}".format(
                        container_name, container, sub_ret
                    )
            else:
                error = sub_ret
                if not error:
                    error = "unknown error (no return)"
            if error:
                ret["result"] = False
                serrs.append(error)
            else:
                container["container_name"] = name
                containers.append(container)
                done.append(container)

        # marking ping status as True only and only if we have at
        # least provisioned one container
        ret["ping_status"] = bool(len(done))

        # for all provisioned containers, last job is to verify
        # - the key status
        # - we can reach them
        for container in done:
            # explicitly check and update
            # the minion key/pair stored on the master
            container_name = container["container_name"]
            key = os.path.join(__opts__["pki_dir"], "minions", container_name)
            if explicit_auth:
                fcontent = ""
                if os.path.exists(key):
                    with salt.utils.files.fopen(key) as fic:
                        fcontent = salt.utils.stringutils.to_unicode(fic.read()).strip()
                pub_key = salt.utils.stringutils.to_unicode(pub_key)
                if pub_key.strip() != fcontent:
                    with salt.utils.files.fopen(key, "w") as fic:
                        fic.write(salt.utils.stringutils.to_str(pub_key))
                        fic.flush()
            mid = j_ret.get("mid", None)
            if not mid:
                continue

            def testping(**kw):
                mid_ = kw["mid"]
                ping = client.cmd(mid_, "test.ping", timeout=20)
                time.sleep(1)
                if ping:
                    return "OK"
                raise Exception(f"Unresponsive {mid_}")

            ping = salt.utils.cloud.wait_for_fun(testping, timeout=21, mid=mid)
            if ping != "OK":
                ret["ping_status"] = False
                ret["result"] = False

        # if no lxc detected as touched (either inited or verified)
        # we result to False
        if not done:
            ret["result"] = False
        if not quiet:
            __jid_event__.fire_event({"message": ret}, "progress")
        return ret


def cloud_init(names, host=None, quiet=False, **kwargs):
    """
    Wrapper for using lxc.init in saltcloud compatibility mode

    names
        Name of the containers, supports a single name or a comma delimited
        list of names.

    host
        Minion to start the container on. Required.

    path
        path to the container parent
        default: /var/lib/lxc (system default)

        .. versionadded:: 2015.8.0

    saltcloud_mode
        init the container with the saltcloud opts format instead
    """
    if quiet:
        log.warning("'quiet' argument is being deprecated. Please migrate to --quiet")
    return __salt__["lxc.init"](
        names=names, host=host, saltcloud_mode=True, quiet=quiet, **kwargs
    )


def _list_iter(host=None, path=None):
    """
    Return a generator iterating over hosts

    path
        path to the container parent
        default: /var/lib/lxc (system default)

        .. versionadded:: 2015.8.0
    """
    tgt = host or "*"
    with salt.client.get_local_client(__opts__["conf_file"]) as client:
        for container_info in client.cmd_iter(tgt, "lxc.list", kwarg={"path": path}):
            if not container_info:
                continue
            if not isinstance(container_info, dict):
                continue
            chunk = {}
            id_ = next(iter(container_info.keys()))
            if host and host != id_:
                continue
            if not isinstance(container_info[id_], dict):
                continue
            if "ret" not in container_info[id_]:
                continue
            if not isinstance(container_info[id_]["ret"], dict):
                continue
            chunk[id_] = container_info[id_]["ret"]
            yield chunk


def list_(host=None, quiet=False, path=None):
    """
    List defined containers (running, stopped, and frozen) for the named
    (or all) host(s).

    path
        path to the container parent
        default: /var/lib/lxc (system default)

        .. versionadded:: 2015.8.0

    .. code-block:: bash

        salt-run lxc.list [host=minion_id]
    """
    it = _list_iter(host, path=path)
    ret = {}
    for chunk in it:
        ret.update(chunk)
        if not quiet:
            __jid_event__.fire_event(
                {"data": chunk, "outputter": "lxc_list"}, "progress"
            )
    return ret


def purge(name, delete_key=True, quiet=False, path=None):
    """
    Purge the named container and delete its minion key if present.
    WARNING: Destroys all data associated with the container.

    path
        path to the container parent
        default: /var/lib/lxc (system default)

        .. versionadded:: 2015.8.0

    .. code-block:: bash

        salt-run lxc.purge name
    """
    data = _do_names(name, "destroy", path=path)
    if data is False:
        return data

    if delete_key:
        with salt.key.Key(__opts__) as skey:
            skey.delete_key(name)

    if data is None:
        return

    if not quiet:
        __jid_event__.fire_event({"data": data, "outputter": "lxc_purge"}, "progress")
    return data


def start(name, quiet=False, path=None):
    """
    Start the named container.

    path
        path to the container parent
        default: /var/lib/lxc (system default)

        .. versionadded:: 2015.8.0

    .. code-block:: bash

        salt-run lxc.start name
    """
    data = _do_names(name, "start", path=path)
    if data and not quiet:
        __jid_event__.fire_event({"data": data, "outputter": "lxc_start"}, "progress")
    return data


def stop(name, quiet=False, path=None):
    """
    Stop the named container.

    path
        path to the container parent
        default: /var/lib/lxc (system default)

        .. versionadded:: 2015.8.0

    .. code-block:: bash

        salt-run lxc.stop name
    """
    data = _do_names(name, "stop", path=path)
    if data and not quiet:
        __jid_event__.fire_event(
            {"data": data, "outputter": "lxc_force_off"}, "progress"
        )
    return data


def freeze(name, quiet=False, path=None):
    """
    Freeze the named container

    path
        path to the container parent
        default: /var/lib/lxc (system default)

        .. versionadded:: 2015.8.0

    .. code-block:: bash

        salt-run lxc.freeze name
    """
    data = _do_names(name, "freeze")
    if data and not quiet:
        __jid_event__.fire_event({"data": data, "outputter": "lxc_pause"}, "progress")
    return data


def unfreeze(name, quiet=False, path=None):
    """
    Unfreeze the named container

    path
        path to the container parent
        default: /var/lib/lxc (system default)

        .. versionadded:: 2015.8.0

    .. code-block:: bash

        salt-run lxc.unfreeze name
    """
    data = _do_names(name, "unfreeze", path=path)
    if data and not quiet:
        __jid_event__.fire_event({"data": data, "outputter": "lxc_resume"}, "progress")
    return data


def info(name, quiet=False, path=None):
    """
    Returns information about a container.

    path
        path to the container parent
        default: /var/lib/lxc (system default)

        .. versionadded:: 2015.8.0

    .. code-block:: bash

        salt-run lxc.info name
    """
    data = _do_names(name, "info", path=path)
    if data and not quiet:
        __jid_event__.fire_event({"data": data, "outputter": "lxc_info"}, "progress")
    return data



import base64
import copy
import logging

from requests.exceptions import ConnectionError

import salt.cache
import salt.crypt
import salt.exceptions
import salt.modules.publish
import salt.modules.saltutil
import salt.utils.context
import salt.utils.data
import salt.utils.dictupdate
import salt.utils.json
import salt.utils.vault.api as vapi
import salt.utils.vault.auth as vauth
import salt.utils.vault.cache as vcache
import salt.utils.vault.client as vclient
import salt.utils.vault.helpers as hlp
import salt.utils.vault.kv as vkv
import salt.utils.vault.leases as vleases
import salt.utils.versions
from salt.defaults import NOT_SET
from salt.utils.vault.exceptions import (
    VaultAuthExpired,
    VaultConfigExpired,
    VaultException,
    VaultPermissionDeniedError,
    VaultUnwrapException,
)

log = logging.getLogger(__name__)
logging.getLogger("requests").setLevel(logging.WARNING)


TOKEN_CKEY = "__token"
CLIENT_CKEY = "_vault_authd_client"


def get_authd_client(opts, context, force_local=False, get_config=False):
    """
    Returns an AuthenticatedVaultClient that is valid for at least one query.
    """

    def try_build():
        client = config = None
        retry = False
        try:
            client, config = _build_authd_client(opts, context, force_local=force_local)
        except (VaultConfigExpired, VaultPermissionDeniedError, ConnectionError):
            clear_cache(opts, context, connection=True, force_local=force_local)
            retry = True
        except VaultUnwrapException as err:
            # ensure to notify about potential intrusion attempt
            _get_event(opts)(tag="vault/security/unwrapping/error", data=err.event_data)
            raise
        return client, config, retry

    cbank = vcache._get_cache_bank(opts, force_local=force_local)
    retry = False
    client = config = None

    # First, check if an already initialized instance is available
    # and still valid
    if cbank in context and CLIENT_CKEY in context[cbank]:
        log.debug("Fetching client instance and config from context")
        client, config = context[cbank][CLIENT_CKEY]
        if not client.token_valid(remote=False):
            log.debug("Cached client instance was invalid")
            client = config = None
            context[cbank].pop(CLIENT_CKEY)

    # Otherwise, try to build one from possibly cached data
    if client is None or config is None:
        try:
            client, config, retry = try_build()
        except VaultAuthExpired:
            clear_cache(opts, context, session=True, force_local=force_local)
            client, config, retry = try_build()

    # Check if the token needs to be and can be renewed.
    # Since this needs to check the possibly active session and does not care
    # about valid secret IDs etc, we need to inspect the actual token.
    if (
        not retry
        and config["auth"]["token_lifecycle"]["renew_increment"] is not False
        and client.auth.get_token().is_renewable()
        and not client.auth.get_token().is_valid(
            config["auth"]["token_lifecycle"]["minimum_ttl"]
        )
    ):
        log.debug("Renewing token")
        client.token_renew(
            increment=config["auth"]["token_lifecycle"]["renew_increment"]
        )

    # Check if the current token could not be renewed for a sufficient amount of time.
    if not retry and not client.token_valid(
        config["auth"]["token_lifecycle"]["minimum_ttl"] or 0, remote=False
    ):
        clear_cache(opts, context, session=True, force_local=force_local)
        client, config, retry = try_build()

    if retry:
        log.debug("Requesting new authentication credentials")
        try:
            client, config = _build_authd_client(opts, context, force_local=force_local)
        except VaultUnwrapException as err:
            _get_event(opts)(tag="vault/security/unwrapping/error", data=err.event_data)
            raise
        if not client.token_valid(
            config["auth"]["token_lifecycle"]["minimum_ttl"] or 0, remote=False
        ):
            if not config["auth"]["token_lifecycle"]["minimum_ttl"]:
                raise VaultException(
                    "Could not build valid client. This is most likely a bug."
                )
            log.warning(
                "Configuration error: auth:token_lifecycle:minimum_ttl cannot be "
                "honored because fresh tokens are issued with less ttl. Continuing anyways."
            )

    if cbank not in context:
        context[cbank] = {}
    context[cbank][CLIENT_CKEY] = (client, config)

    if get_config:
        return client, config
    return client


def clear_cache(
    opts, context, ckey=None, connection=True, session=False, force_local=False
):
    """
    Clears the Vault cache.
    Will ensure the current token and associated leases are revoked
    by default.

    It is organized in a hierarchy: ``/vault/connection/session/leases``.
    (*italics* mark data that is only cached when receiving configuration from a master)

    ``connection`` contains KV metadata (by default), *configuration* and *(AppRole) auth credentials*.
    ``session`` contains the currently active token.
    ``leases`` contains leases issued to the currently active token like database credentials.

    A master keeps a separate instance of the above per minion
    in ``minions/<minion_id>``.

    opts
        Pass ``__opts__``.

    context
        Pass ``__context__``.

    ckey
        Only clear this cache key instead of the whole cache bank.

    connection
        Only clear the cached data scoped to a connection. This includes
        configuration, auth credentials, the currently active auth token
        as well as leases and KV metadata (by default). Defaults to true.
        Set this to false to clear all Vault caches.

    session
        Only clear the cached data scoped to a session. This only includes
        leases and the currently active auth token, but not configuration
        or (AppRole) auth credentials. Defaults to false.
        Setting this to true will keep the connection cache, regardless
        of ``connection``.

    force_local
        Required on the master when the runner is issuing credentials during
        pillar compilation. Instructs the cache to use the ``/vault`` cache bank,
        regardless of determined run type. Defaults to false and should not
        be set by anything other than the runner.
    """
    cbank = vcache._get_cache_bank(
        opts, connection=connection, session=session, force_local=force_local
    )
    client, config = _build_revocation_client(opts, context, force_local=force_local)
    if (
        not ckey
        or (not (connection or session) and ckey == "connection")
        or (session and ckey == TOKEN_CKEY)
        or ((connection and not session) and ckey == "config")
    ):
        client, config = _build_revocation_client(
            opts, context, force_local=force_local
        )
        # config and client will both be None if the cached data is invalid
        if config:
            try:
                # Don't revoke the only token that is available to us
                if config["auth"]["method"] != "token" or not (
                    force_local
                    or hlp._get_salt_run_type(opts)
                    in (hlp.SALT_RUNTYPE_MASTER, hlp.SALT_RUNTYPE_MINION_LOCAL)
                ):
                    if config["cache"]["clear_attempt_revocation"]:
                        delta = config["cache"]["clear_attempt_revocation"]
                        if delta is True:
                            delta = 1
                        client.token_revoke(delta)
                    if (
                        config["cache"]["expire_events"]
                        and not force_local
                        and hlp._get_salt_run_type(opts)
                        not in [
                            hlp.SALT_RUNTYPE_MASTER_IMPERSONATING,
                            hlp.SALT_RUNTYPE_MASTER_PEER_RUN,
                        ]
                    ):
                        scope = cbank.split("/")[-1]
                        _get_event(opts)(  # pylint: disable=no-value-for-parameter
                            tag=f"vault/cache/{scope}/clear"
                        )
            except Exception as err:  # pylint: disable=broad-except
                log.error(
                    "Failed to revoke token or send event before clearing cache:\n%s: %s",
                    type(err).__name__,
                    err,
                )

    if cbank in context:
        if ckey is None:
            context.pop(cbank)
        else:
            context[cbank].pop(ckey, None)
            if connection and not session:
                # Ensure the active client gets recreated after altering the connection cache
                context[cbank].pop(CLIENT_CKEY, None)

    # also remove sub-banks from context to mimic cache behavior
    if ckey is None:
        for bank in list(context):
            if bank.startswith(cbank):
                context.pop(bank)
    cache = salt.cache.factory(opts)
    if cache.contains(cbank, ckey):
        return cache.flush(cbank, ckey)

    # In case the cache driver was overridden for the Vault integration
    local_opts = copy.copy(opts)
    opts["cache"] = "localfs"
    cache = salt.cache.factory(local_opts)
    return cache.flush(cbank, ckey)


def update_config(opts, context, keep_session=False):
    """
    Attempt to update the cached configuration without
    clearing the currently active Vault session.

    opts
        Pass __opts__.

    context
        Pass __context__.

    keep_session
        Only update configuration that can be updated without
        creating a new login session.
        If this is false, still tries to keep the active session,
        but might clear it if the server configuration has changed
        significantly.
        Defaults to False.
    """
    if hlp._get_salt_run_type(opts) in [
        hlp.SALT_RUNTYPE_MASTER,
        hlp.SALT_RUNTYPE_MINION_LOCAL,
    ]:
        # local configuration is not cached
        return True
    connection_cbank = vcache._get_cache_bank(opts)
    try:
        _get_connection_config(connection_cbank, opts, context, update=True)
        return True
    except VaultConfigExpired:
        pass
    if keep_session:
        return False
    clear_cache(opts, context, connection=True)
    get_authd_client(opts, context)
    return True


def _build_authd_client(opts, context, force_local=False):
    connection_cbank = vcache._get_cache_bank(opts, force_local=force_local)
    config, embedded_token, unauthd_client = _get_connection_config(
        connection_cbank, opts, context, force_local=force_local
    )
    # Tokens are cached in a distinct scope to enable cache per session
    session_cbank = vcache._get_cache_bank(opts, force_local=force_local, session=True)
    cache_ttl = (
        config["cache"]["secret"] if config["cache"]["secret"] != "ttl" else None
    )
    token_cache = vcache.VaultAuthCache(
        context,
        session_cbank,
        TOKEN_CKEY,
        vleases.VaultToken,
        cache_backend=vcache._get_cache_backend(config, opts),
        ttl=cache_ttl,
        flush_exception=VaultAuthExpired,
    )

    client = None

    if config["auth"]["method"] == "approle":
        secret_id = config["auth"]["secret_id"] or None
        cached_token = token_cache.get(10)
        secret_id_cache = None
        if secret_id:
            secret_id_cache = vcache.VaultAuthCache(
                context,
                connection_cbank,
                "secret_id",
                vleases.VaultSecretId,
                cache_backend=vcache._get_cache_backend(config, opts),
                ttl=cache_ttl,
            )
            secret_id = secret_id_cache.get()
            # Only fetch secret ID if there is no cached valid token
            if cached_token is None and secret_id is None:
                secret_id = _fetch_secret_id(
                    config,
                    opts,
                    secret_id_cache,
                    unauthd_client,
                    force_local=force_local,
                )
            if secret_id is None:
                # If the auth config is sourced locally, ensure the
                # SecretID is known regardless whether we have a valid token.
                # For remote sources, we would needlessly request one, so don't.
                if (
                    hlp._get_salt_run_type(opts)
                    in [hlp.SALT_RUNTYPE_MASTER, hlp.SALT_RUNTYPE_MINION_LOCAL]
                    or force_local
                ):
                    secret_id = _fetch_secret_id(
                        config,
                        opts,
                        secret_id_cache,
                        unauthd_client,
                        force_local=force_local,
                    )
                else:
                    secret_id = vauth.InvalidVaultSecretId()
        role_id = config["auth"]["role_id"]
        # this happens with wrapped response merging
        if isinstance(role_id, dict):
            role_id = role_id["role_id"]
        approle = vauth.VaultAppRole(role_id, secret_id)
        token_auth = vauth.VaultTokenAuth(cache=token_cache)
        auth = vauth.VaultAppRoleAuth(
            approle,
            unauthd_client,
            mount=config["auth"]["approle_mount"],
            cache=secret_id_cache,
            token_store=token_auth,
        )
        client = vclient.AuthenticatedVaultClient(
            auth, session=unauthd_client.session, **config["server"]
        )
    elif config["auth"]["method"] in ["token", "wrapped_token"]:
        token = _fetch_token(
            config,
            opts,
            token_cache,
            unauthd_client,
            force_local=force_local,
            embedded_token=embedded_token,
        )
        auth = vauth.VaultTokenAuth(token=token, cache=token_cache)
        client = vclient.AuthenticatedVaultClient(
            auth, session=unauthd_client.session, **config["server"]
        )

    if client is not None:
        return client, config
    raise salt.exceptions.SaltException("Connection configuration is invalid.")


def _build_revocation_client(opts, context, force_local=False):
    """
    Tries to build an AuthenticatedVaultClient solely from caches.
    This client is used to revoke all leases before forgetting about them.
    """
    connection_cbank = vcache._get_cache_bank(opts, force_local=force_local)
    # Disregard a possibly returned locally configured token since
    # it is cached with metadata if it has been used. Also, we do not want
    # to revoke statically configured tokens anyways.
    config, _, unauthd_client = _get_connection_config(
        connection_cbank, opts, context, force_local=force_local, pre_flush=True
    )
    if config is None:
        return None, None

    # Tokens are cached in a distinct scope to enable cache per session
    session_cbank = vcache._get_cache_bank(opts, force_local=force_local, session=True)
    token_cache = vcache.VaultAuthCache(
        context,
        session_cbank,
        TOKEN_CKEY,
        vleases.VaultToken,
        cache_backend=vcache._get_cache_backend(config, opts),
    )

    token = token_cache.get(flush=False)

    if token is None:
        return None, None
    auth = vauth.VaultTokenAuth(token=token, cache=token_cache)
    client = vclient.AuthenticatedVaultClient(auth, **config["server"])
    return client, config


def _get_connection_config(
    cbank, opts, context, force_local=False, pre_flush=False, update=False
):
    if (
        hlp._get_salt_run_type(opts)
        in [hlp.SALT_RUNTYPE_MASTER, hlp.SALT_RUNTYPE_MINION_LOCAL]
        or force_local
    ):
        # only cache config fetched from remote
        return _use_local_config(opts)

    if pre_flush and update:
        raise VaultException("`pre_flush` and `update` are mutually exclusive")

    log.debug("Using Vault server connection configuration from remote.")
    config_cache = vcache._get_config_cache(opts, context, cbank)
    if pre_flush:
        # ensure any cached data is tried when building a client for revocation
        config_cache.ttl = None
    # In case cached data is available, this takes care of bubbling up
    # an exception indicating all connection-scoped data should be flushed
    # if the config is outdated.
    config = config_cache.get()
    if config is not None and not update:
        log.debug("Using cached Vault server connection configuration.")
        return config, None, vclient.VaultClient(**config["server"])

    if pre_flush:
        # used when building a client that revokes leases before clearing cache
        return None, None, None

    log.debug("Using new Vault server connection configuration.")
    try:
        issue_params = parse_config(opts.get("vault", {}), validate=False)[
            "issue_params"
        ]
        new_config, unwrap_client = _query_master(
            "get_config",
            opts,
            issue_params=issue_params or None,
            config_only=update,
        )
    except VaultConfigExpired as err:
        # Make sure to still work with old peer_run configuration
        if "Peer runner return was empty" not in err.message or update:
            raise
        log.warning(
            "Got empty response to Vault config request. Falling back to vault.generate_token. "
            "Please update your master peer_run configuration."
        )
        new_config, unwrap_client = _query_master(
            "generate_token",
            opts,
            ttl=issue_params.get("explicit_max_ttl"),
            uses=issue_params.get("num_uses"),
            upgrade_request=True,
        )
    new_config = parse_config(new_config, opts=opts, require_token=not update)
    # do not couple token cache with configuration cache
    embedded_token = new_config["auth"].pop("token", None)
    new_config = {
        "auth": new_config["auth"],
        "cache": new_config["cache"],
        "server": new_config["server"],
    }
    if update and config:
        if new_config["server"] != config["server"]:
            raise VaultConfigExpired()
        if new_config["auth"]["method"] != config["auth"]["method"]:
            raise VaultConfigExpired()
        if new_config["auth"]["method"] == "approle" and (
            new_config["auth"]["role_id"] != config["auth"]["role_id"]
            or new_config["auth"]["secret_id"] is not config["auth"]["secret_id"]
        ):
            # enabling/disabling response wrapping will trigger this as well,
            # but that's fine
            raise VaultConfigExpired()
        if new_config["cache"]["backend"] != config["cache"]["backend"]:
            raise VaultConfigExpired()
        config_cache.flush(cbank=False)

    config_cache.store(new_config)
    return new_config, embedded_token, unwrap_client


def _use_local_config(opts):
    log.debug("Using Vault connection details from local config.")
    config = parse_config(opts.get("vault", {}))
    embedded_token = config["auth"].pop("token", None)
    return (
        {
            "auth": config["auth"],
            "cache": config["cache"],
            "server": config["server"],
        },
        embedded_token,
        vclient.VaultClient(**config["server"]),
    )


def _fetch_secret_id(config, opts, secret_id_cache, unwrap_client, force_local=False):
    def cache_or_fetch(config, opts, secret_id_cache, unwrap_client):
        secret_id = secret_id_cache.get()
        if secret_id is not None:
            return secret_id

        log.debug("Fetching new Vault AppRole secret ID.")
        secret_id, _ = _query_master(
            "generate_secret_id",
            opts,
            unwrap_client=unwrap_client,
            unwrap_expected_creation_path=vclient._get_expected_creation_path(
                "secret_id", config
            ),
            issue_params=parse_config(opts.get("vault", {}), validate=False)[
                "issue_params"
            ]
            or None,
        )
        secret_id = vleases.VaultSecretId(**secret_id["data"])
        # Do not cache single-use secret IDs
        if secret_id.num_uses != 1:
            secret_id_cache.store(secret_id)
        return secret_id

    if (
        hlp._get_salt_run_type(opts)
        in [hlp.SALT_RUNTYPE_MASTER, hlp.SALT_RUNTYPE_MINION_LOCAL]
        or force_local
    ):
        secret_id = config["auth"]["secret_id"]
        if isinstance(secret_id, dict):
            if secret_id.get("wrap_info"):
                secret_id = unwrap_client.unwrap(
                    secret_id["wrap_info"]["token"],
                    expected_creation_path=vclient._get_expected_creation_path(
                        "secret_id", config
                    ),
                )
                secret_id = secret_id["data"]
            return vauth.LocalVaultSecretId(**secret_id)
        if secret_id:
            # assume locally configured secret_ids do not expire
            return vauth.LocalVaultSecretId(
                secret_id=config["auth"]["secret_id"],
                secret_id_ttl=0,
                secret_id_num_uses=0,
            )
        # When secret_id is falsey, the approle does not require secret IDs,
        # hence a call to this function is superfluous
        raise salt.exceptions.SaltException("This code path should not be hit at all.")

    log.debug("Using secret_id issued by master.")
    return cache_or_fetch(config, opts, secret_id_cache, unwrap_client)


def _fetch_token(
    config, opts, token_cache, unwrap_client, force_local=False, embedded_token=None
):
    def cache_or_fetch(config, opts, token_cache, unwrap_client, embedded_token):
        token = token_cache.get(10)
        if token is not None:
            log.debug("Using cached token.")
            return token

        if isinstance(embedded_token, dict):
            token = vleases.VaultToken(**embedded_token)

        if not isinstance(token, vleases.VaultToken) or not token.is_valid(10):
            log.debug("Fetching new Vault token.")
            token, _ = _query_master(
                "generate_new_token",
                opts,
                unwrap_client=unwrap_client,
                unwrap_expected_creation_path=vclient._get_expected_creation_path(
                    "token", config
                ),
                issue_params=parse_config(opts.get("vault", {}), validate=False)[
                    "issue_params"
                ]
                or None,
            )
            token = vleases.VaultToken(**token["auth"])

        # do not cache single-use tokens
        if token.num_uses != 1:
            token_cache.store(token)
        return token

    if (
        hlp._get_salt_run_type(opts)
        in [hlp.SALT_RUNTYPE_MASTER, hlp.SALT_RUNTYPE_MINION_LOCAL]
        or force_local
    ):
        token = None
        if isinstance(embedded_token, dict):
            if embedded_token.get("wrap_info"):
                embedded_token = unwrap_client.unwrap(
                    embedded_token["wrap_info"]["token"],
                    expected_creation_path=vclient._get_expected_creation_path(
                        "token", config
                    ),
                )["auth"]
            token = vleases.VaultToken(**embedded_token)
        elif config["auth"]["method"] == "wrapped_token":
            embedded_token = unwrap_client.unwrap(
                embedded_token,
                expected_creation_path=vclient._get_expected_creation_path(
                    "token", config
                ),
            )["auth"]
            token = vleases.VaultToken(**embedded_token)
        elif embedded_token is not None:
            # if the embedded plain token info has been cached before, don't repeat
            # the query unnecessarily
            token = token_cache.get()
            if token is None or embedded_token != str(token):
                # lookup and verify raw token
                token_info = unwrap_client.token_lookup(embedded_token, raw=True)
                if token_info.status_code != 200:
                    raise VaultException(
                        "Configured token cannot be verified. It is most likely expired or invalid."
                    )
                token_meta = token_info.json()["data"]
                token = vleases.VaultToken(
                    lease_id=embedded_token,
                    lease_duration=token_meta["ttl"],
                    **token_meta,
                )
                token_cache.store(token)
        if token is not None:
            return token
        raise VaultException("Invalid configuration, missing token.")

    log.debug("Using token generated by master.")
    return cache_or_fetch(config, opts, token_cache, unwrap_client, embedded_token)


def _query_master(
    func,
    opts,
    unwrap_client=None,
    unwrap_expected_creation_path=None,
    **kwargs,
):
    def check_result(
        result,
        unwrap_client=None,
        unwrap_expected_creation_path=None,
    ):
        if not result:
            log.error(
                "Failed to get Vault connection from master! No result returned - "
                "does the peer runner publish configuration include `vault.%s`?",
                func,
            )
            # Expire configuration in case this is the result of an auth method change.
            raise VaultConfigExpired(
                f"Peer runner return was empty. Make sure {func} is listed in the master peer_run config."
            )
        if not isinstance(result, dict):
            log.error(
                "Failed to get Vault connection from master! Response is not a dict: %s",
                result,
            )
            raise salt.exceptions.CommandExecutionError(result)
        if "error" in result:
            log.error(
                "Failed to get Vault connection from master! An error was returned: %s",
                result["error"],
            )
            if result.get("expire_cache"):
                log.warning("Master returned error and requested cache expiration.")
                raise VaultConfigExpired()
            raise salt.exceptions.CommandExecutionError(result)

        config_expired = False
        expected_server = None

        if result.get("expire_cache", False):
            log.info("Master requested Vault config expiration.")
            config_expired = True

        if "server" in result:
            # Ensure locally overridden verify parameter does not
            # always invalidate cache.
            reported_server = parse_config(result["server"], validate=False, opts=opts)[
                "server"
            ]
            result.update({"server": reported_server})

        if unwrap_client is not None:
            expected_server = unwrap_client.get_config()

        if expected_server is not None and result.get("server") != expected_server:
            log.info(
                "Mismatch of cached and reported server data detected. Invalidating cache."
            )
            # make sure to fetch wrapped data anyways for security reasons
            config_expired = True
            unwrap_expected_creation_path = None
            unwrap_client = None

        # This is used to augment some vault responses with data fetched by the master
        # e.g. secret_id_num_uses
        misc_data = result.get("misc_data", {})

        if result.get("wrap_info") or result.get("wrap_info_nested"):
            if unwrap_client is None:
                unwrap_client = vclient.VaultClient(**result["server"])

            for key in [""] + result.get("wrap_info_nested", []):
                if key:
                    wrapped = salt.utils.data.traverse_dict(result, key)
                else:
                    wrapped = result
                if not wrapped or "wrap_info" not in wrapped:
                    continue
                wrapped_response = vleases.VaultWrappedResponse(**wrapped["wrap_info"])
                try:
                    unwrapped_response = unwrap_client.unwrap(
                        wrapped_response,
                        expected_creation_path=unwrap_expected_creation_path,
                    )
                except VaultUnwrapException as err:
                    err.event_data.update({"func": f"vault.{func}"})
                    raise
                if key:
                    salt.utils.dictupdate.set_dict_key_value(
                        result,
                        key,
                        unwrapped_response.get("auth")
                        or unwrapped_response.get("data"),
                    )
                else:
                    if unwrapped_response.get("auth"):
                        result.update({"auth": unwrapped_response["auth"]})
                    if unwrapped_response.get("data"):
                        result.update({"data": unwrapped_response["data"]})

        if config_expired:
            raise VaultConfigExpired()

        for key, val in misc_data.items():
            tgt = "data" if result.get("data") is not None else "auth"
            if (
                salt.utils.data.traverse_dict_and_list(result, f"{tgt}:{key}", NOT_SET)
                == NOT_SET
            ):
                salt.utils.dictupdate.set_dict_key_value(
                    result,
                    f"{tgt}:{key}",
                    val,
                )

        result.pop("wrap_info", None)
        result.pop("wrap_info_nested", None)
        result.pop("misc_data", None)
        return result, unwrap_client

    minion_id = opts["grains"]["id"]
    pki_dir = opts["pki_dir"]

    # When rendering pillars, the module executes on the master, but the token
    # should be issued for the minion, so that the correct policies are applied
    if opts.get("__role", "minion") == "minion":
        private_key = f"{pki_dir}/minion.pem"
        log.debug(
            "Running on minion, signing request `vault.%s` with key %s",
            func,
            private_key,
        )
        signature = base64.b64encode(salt.crypt.sign_message(private_key, minion_id))
        arg = [
            ("minion_id", minion_id),
            ("signature", signature),
            ("impersonated_by_master", False),
        ] + list(kwargs.items())
        with salt.utils.context.func_globals_inject(
            salt.modules.publish.runner, __opts__=opts
        ):
            result = salt.modules.publish.runner(
                f"vault.{func}", arg=[{"__kwarg__": True, k: v} for k, v in arg]
            )
    else:
        private_key = f"{pki_dir}/master.pem"
        log.debug(
            "Running on master, signing request `vault.%s` for %s with key %s",
            func,
            minion_id,
            private_key,
        )
        signature = base64.b64encode(salt.crypt.sign_message(private_key, minion_id))
        with salt.utils.context.func_globals_inject(
            salt.modules.saltutil.runner, __opts__=opts
        ):
            result = salt.modules.saltutil.runner(
                f"vault.{func}",
                minion_id=minion_id,
                signature=signature,
                impersonated_by_master=True,
                **kwargs,
            )
    return check_result(
        result,
        unwrap_client=unwrap_client,
        unwrap_expected_creation_path=unwrap_expected_creation_path,
    )


def _get_event(opts):
    event = salt.utils.event.get_event(
        opts.get("__role", "minion"), sock_dir=opts["sock_dir"], opts=opts, listen=False
    )

    if opts.get("__role", "minion") == "minion":
        return event.fire_master
    return event.fire_event


def get_kv(opts, context, get_config=False):
    """
    Return an instance of VaultKV, which can be used
    to interact with the ``kv`` backend.
    """
    client, config = get_authd_client(opts, context, get_config=True)
    ttl = None
    connection = True
    if config["cache"]["kv_metadata"] != "connection":
        ttl = config["cache"]["kv_metadata"]
        connection = False
    cbank = vcache._get_cache_bank(opts, connection=connection)
    ckey = "secret_path_metadata"
    metadata_cache = vcache.VaultCache(
        context,
        cbank,
        ckey,
        cache_backend=vcache._get_cache_backend(config, opts),
        ttl=ttl,
    )
    kv = vkv.VaultKV(client, metadata_cache)
    if get_config:
        return kv, config
    return kv


def get_lease_store(opts, context, get_config=False):
    """
    Return an instance of LeaseStore, which can be used
    to cache leases and handle operations like renewals and revocations.
    """
    client, config = get_authd_client(opts, context, get_config=True)
    session_cbank = vcache._get_cache_bank(opts, session=True)
    expire_events = None
    if config["cache"]["expire_events"]:
        expire_events = _get_event(opts)
    lease_cache = vcache.VaultLeaseCache(
        context,
        session_cbank + "/leases",
        cache_backend=vcache._get_cache_backend(config, opts),
        expire_events=expire_events,
    )
    store = vleases.LeaseStore(client, lease_cache, expire_events=expire_events)
    if get_config:
        return store, config
    return store


def get_approle_api(opts, context, force_local=False, get_config=False):
    """
    Return an instance of AppRoleApi containing an AuthenticatedVaultClient.
    """
    client, config = get_authd_client(
        opts, context, force_local=force_local, get_config=True
    )
    api = vapi.AppRoleApi(client)
    if get_config:
        return api, config
    return api


def get_identity_api(opts, context, force_local=False, get_config=False):
    """
    Return an instance of IdentityApi containing an AuthenticatedVaultClient.
    """
    client, config = get_authd_client(
        opts, context, force_local=force_local, get_config=True
    )
    api = vapi.IdentityApi(client)
    if get_config:
        return api, config
    return api


def parse_config(config, validate=True, opts=None, require_token=True):
    """
    Returns a vault configuration dictionary that has all
    keys with defaults. Checks if required data is available.
    """
    default_config = {
        "auth": {
            "approle_mount": "approle",
            "approle_name": "salt-master",
            "method": "token",
            "secret_id": None,
            "token_lifecycle": {
                "minimum_ttl": 10,
                "renew_increment": None,
            },
        },
        "cache": {
            "backend": "session",
            "clear_attempt_revocation": 60,
            "clear_on_unauthorized": True,
            "config": 3600,
            "expire_events": False,
            "kv_metadata": "connection",
            "secret": "ttl",
        },
        "issue": {
            "allow_minion_override_params": False,
            "type": "token",
            "approle": {
                "mount": "salt-minions",
                "params": {
                    "bind_secret_id": True,
                    "secret_id_num_uses": 1,
                    "secret_id_ttl": 60,
                    "token_explicit_max_ttl": 60,
                    "token_num_uses": 10,
                },
            },
            "token": {
                "role_name": None,
                "params": {
                    "explicit_max_ttl": None,
                    "num_uses": 1,
                },
            },
            "wrap": "30s",
        },
        "issue_params": {},
        "metadata": {
            "entity": {
                "minion-id": "{minion}",
            },
            "secret": {
                "saltstack-jid": "{jid}",
                "saltstack-minion": "{minion}",
                "saltstack-user": "{user}",
            },
        },
        "policies": {
            "assign": [
                "saltstack/minions",
                "saltstack/{minion}",
            ],
            "cache_time": 60,
            "refresh_pillar": None,
        },
        "server": {
            "namespace": None,
            "verify": None,
        },
    }
    try:
        # Policy generation has params, the new config groups them together.
        if isinstance(config.get("policies", {}), list):
            config["policies"] = {"assign": config.pop("policies")}
        merged = salt.utils.dictupdate.merge(
            default_config,
            config,
            strategy="smart",
            merge_lists=False,
        )
        # ttl, uses were used as configuration for issuance and minion overrides as well
        # as token meta information. The new configuration splits those semantics.
        for old_token_conf, new_token_conf in [
            ("ttl", "explicit_max_ttl"),
            ("uses", "num_uses"),
        ]:
            if old_token_conf in merged["auth"]:
                merged["issue"]["token"]["params"][new_token_conf] = merged[
                    "issue_params"
                ][new_token_conf] = merged["auth"].pop(old_token_conf)
        # Those were found in the root namespace, but grouping them together
        # makes semantic and practical sense.
        for old_server_conf in ["namespace", "url", "verify"]:
            if old_server_conf in merged:
                merged["server"][old_server_conf] = merged.pop(old_server_conf)
        if "role_name" in merged:
            merged["issue"]["token"]["role_name"] = merged.pop("role_name")
        if "token_backend" in merged["auth"]:
            merged["cache"]["backend"] = merged["auth"].pop("token_backend")
        if "allow_minion_override" in merged["auth"]:
            merged["issue"]["allow_minion_override_params"] = merged["auth"].pop(
                "allow_minion_override"
            )
        if opts is not None and "vault" in opts:
            local_config = opts["vault"]
            # Respect locally configured verify parameter
            if local_config.get("verify", NOT_SET) != NOT_SET:
                merged["server"]["verify"] = local_config["verify"]
            elif local_config.get("server", {}).get("verify", NOT_SET) != NOT_SET:
                merged["server"]["verify"] = local_config["server"]["verify"]
            # same for token_lifecycle
            if local_config.get("auth", {}).get("token_lifecycle"):
                merged["auth"]["token_lifecycle"] = local_config["auth"][
                    "token_lifecycle"
                ]

        if not validate:
            return merged

        if merged["auth"]["method"] == "approle":
            if "role_id" not in merged["auth"]:
                raise AssertionError("auth:role_id is required for approle auth")
        elif merged["auth"]["method"] == "token":
            if require_token and "token" not in merged["auth"]:
                raise AssertionError("auth:token is required for token auth")
        else:
            raise AssertionError(
                f"`{merged['auth']['method']}` is not a valid auth method."
            )

        if "url" not in merged["server"]:
            raise AssertionError("server:url is required")
    except AssertionError as err:
        raise salt.exceptions.InvalidConfigError(
            f"Invalid vault configuration: {err}"
        ) from err
    return merged



"""uapi functions"""
from typing import Union
from ._base import CpAPIBase
from ._errors import CpAPIErrorMsg


class Uapi(CpAPIBase):
    """uapi functions"""

    __module__ = 'cpapis'

    def __init__(self):
        if self._can_exec('/usr/bin/uapi'):
            super().__init__('/usr/bin/uapi')
        else:
            super().__init__('/usr/local/cpanel/bin/uapi')

    def __call__(
        self,
        module: str,
        user: Union[str, None] = None,
        args: Union[dict, None] = None,
        timeout: Union[float, None] = None,
        *,
        check: bool = False,
    ):
        """Query cPanel uapi

        Args:
            module: uapi module to use in format x::y
            user: ``--user`` arg to use
            args: key-vals to send to uapi
            timeout: timeout for the uapi command in secs
            check: check .result.status from the result and raise
                an CpAPIErrorMsg on failure. Defaults False.

        Raises:
            CpAPIExecFail: command failed to execute or returned invalid json
            CpAPIErrorMsg: check=True was set and the API reported an error
        """
        ret = self._exec(
            module_args=module.split('::'),
            user=user,
            args=args,
            timeout=timeout,
        )
        if check:
            self.check(ret)
        return ret

    @classmethod
    def check(cls, data: dict) -> None:
        """Parse output common to most uapi functions and raise an
        exception if the function failed

        Args:
            data: result data from a uapi call

        Raises:
            CpAPIErrorMsg: the API reported an error
        """
        try:
            if data['result']['status'] != 1:
                msg = data['result'].get("errors", None)
                if not msg:
                    msg = data['result'].get('messages', None)
                if not msg:
                    msg = "unspecified error"
                if isinstance(msg, list):
                    msg = '. '.join(map(str, msg))
                else:
                    msg = str(msg)
                raise CpAPIErrorMsg(msg=msg, data=data)
        except (TypeError, KeyError) as exc:
            raise CpAPIErrorMsg(
                msg=f"{type(exc).__name__}: {exc}", data=data
            ) from exc

    def create_mysql_db(
        self, user: str, dbname: str, timeout: Union[float, None] = None
    ) -> None:
        """Create a MySQL database

        Args:
            user: cPanel user to assign the database to
            dbname: database name including its prefix
            timeout: timeout for the uapi command in secs

        Raises:
            CpAPIExecFail: command failed to execute or returned invalid json
            CpAPIErrorMsg: the API reported an error
        """
        self(
            module='Mysql::create_database',
            user=user,
            args={'name': dbname},
            check=True,
            timeout=timeout,
        )

    def create_pgsql_db(
        self, user: str, dbname: str, timeout: Union[float, None] = None
    ) -> None:
        """Create a PostgreSQL database

        Args:
            user: cPanel user to assign the database to
            dbname: database name including its prefix
            timeout: timeout for the uapi command in secs

        Raises:
            CpAPIExecFail: command failed to execute or returned invalid json
            CpAPIErrorMsg: the API reported an error
        """
        self(
            module='Postgresql::create_database',
            user=user,
            args={'name': dbname},
            check=True,
            timeout=timeout,
        )



"""
Manage Grafana v4.0 users

.. versionadded:: 2017.7.0

:configuration: This state requires a configuration profile to be configured
    in the minion config, minion pillar, or master config. The module will use
    the 'grafana' key by default, if defined.

    Example configuration using basic authentication:

    .. code-block:: yaml

        grafana:
          grafana_url: http://grafana.localhost
          grafana_user: admin
          grafana_password: admin
          grafana_timeout: 3

    Example configuration using token based authentication:

    .. code-block:: yaml

        grafana:
          grafana_url: http://grafana.localhost
          grafana_token: token
          grafana_timeout: 3

.. code-block:: yaml

    Ensure foobar user is present:
      grafana4_user.present:
        - name: foobar
        - password: mypass
        - email: "foobar@localhost"
        - fullname: Foo Bar
        - is_admin: true
"""

import salt.utils.dictupdate as dictupdate
from salt.utils.dictdiffer import deep_diff


def __virtual__():
    """Only load if grafana4 module is available"""
    if "grafana4.get_user" in __salt__:
        return True
    return (False, "grafana4 module could not be loaded")


def present(
    name, password, email, is_admin=False, fullname=None, theme=None, profile="grafana"
):
    """
    Ensure that a user is present.

    name
        Name of the user.

    password
        Password of the user.

    email
        Email of the user.

    is_admin
        Optional - Set user as admin user. Default: False

    fullname
        Optional - Full name of the user.

    theme
        Optional - Selected theme of the user.

    profile
        Configuration profile used to connect to the Grafana instance.
        Default is 'grafana'.
    """
    if isinstance(profile, str):
        profile = __salt__["config.option"](profile)

    ret = {"name": name, "result": None, "comment": None, "changes": {}}
    user = __salt__["grafana4.get_user"](name, profile)
    create = not user

    if create:
        if __opts__["test"]:
            ret["comment"] = f"User {name} will be created"
            return ret
        __salt__["grafana4.create_user"](
            login=name, password=password, email=email, name=fullname, profile=profile
        )
        user = __salt__["grafana4.get_user"](name, profile)
        ret["changes"]["new"] = user

    user_data = __salt__["grafana4.get_user_data"](user["id"], profile=profile)
    data = _get_json_data(
        login=name, email=email, name=fullname, theme=theme, defaults=user_data
    )
    if data != _get_json_data(
        login=None, email=None, name=None, theme=None, defaults=user_data
    ):
        if __opts__["test"]:
            ret["comment"] = f"User {name} will be updated"
            return ret
        __salt__["grafana4.update_user"](user["id"], profile=profile, **data)
        dictupdate.update(
            ret["changes"],
            deep_diff(user_data, __salt__["grafana4.get_user_data"](user["id"])),
        )

    if user["isAdmin"] != is_admin:
        if __opts__["test"]:
            ret["comment"] = f"User {name} isAdmin status will be updated"
            return ret
        __salt__["grafana4.update_user_permissions"](
            user["id"], isGrafanaAdmin=is_admin, profile=profile
        )
        dictupdate.update(
            ret["changes"],
            deep_diff(user, __salt__["grafana4.get_user"](name, profile)),
        )

    ret["result"] = True
    if create:
        ret["changes"] = ret["changes"]["new"]
        ret["comment"] = f"New user {name} added"
    else:
        if ret["changes"]:
            ret["comment"] = f"User {name} updated"
        else:
            ret["changes"] = {}
            ret["comment"] = f"User {name} already up-to-date"

    return ret


def absent(name, profile="grafana"):
    """
    Ensure that a user is present.

    name
        Name of the user to remove.

    profile
        Configuration profile used to connect to the Grafana instance.
        Default is 'grafana'.
    """
    if isinstance(profile, str):
        profile = __salt__["config.option"](profile)

    ret = {"name": name, "result": None, "comment": None, "changes": {}}
    user = __salt__["grafana4.get_user"](name, profile)

    if user:
        if __opts__["test"]:
            ret["comment"] = f"User {name} will be deleted"
            return ret
        orgs = __salt__["grafana4.get_user_orgs"](user["id"], profile=profile)
        __salt__["grafana4.delete_user"](user["id"], profile=profile)
        for org in orgs:
            if org["name"] == user["email"]:
                # Remove entire Org in the case where auto_assign_org=false:
                # When set to false, new users will automatically cause a new
                # organization to be created for that new user (the org name
                # will be the email)
                __salt__["grafana4.delete_org"](org["orgId"], profile=profile)
            else:
                __salt__["grafana4.delete_user_org"](
                    user["id"], org["orgId"], profile=profile
                )
    else:
        ret["result"] = True
        ret["comment"] = f"User {name} already absent"
        return ret

    ret["result"] = True
    ret["changes"][name] = "Absent"
    ret["comment"] = f"User {name} was deleted"
    return ret


def _get_json_data(defaults=None, **kwargs):
    if defaults is None:
        defaults = {}
    for k, v in kwargs.items():
        if v is None:
            kwargs[k] = defaults.get(k)
    return kwargs



"""Restic Exception Classes"""

from typing import Union
import os
import re
import platform
import logging
from signal import Signals, SIGPIPE
from subprocess import CalledProcessError, CompletedProcess

LOCK_RE = re.compile(r'locked (?:exclusively )?by PID (\d+) on ([\w\.\-]+) by')


class ResticError(Exception):
    """Base class for restic errors

    Args:
        proc (CalledProcessError | CompletedProcess): subprocess result
    """

    __module__ = 'restic'

    def __init__(self, proc: Union[CalledProcessError, CompletedProcess]):
        if isinstance(proc.stderr, bytes):
            stderr = str(proc.stderr, 'utf-8')
        else:
            stderr = proc.stderr
        super().__init__(stderr)
        self.stderr = stderr
        self.returncode = proc.returncode

    def __new__(cls, proc: Union[CalledProcessError, CompletedProcess]):
        if cls is not ResticError:
            return Exception.__new__(cls)
        if isinstance(proc.stderr, bytes):
            stderr = str(proc.stderr, 'utf-8').strip()
        else:
            stderr = proc.stderr.strip()
        if (
            'dial tcp' in stderr
            or 'TLS handshake timeout' in stderr
            or '502 Bad Gateway' in stderr
        ):
            return Exception.__new__(ResticConnError)
        if 'The access key ID you provided does not exist' in stderr:
            return Exception.__new__(ResticAccessError)
        if (
            'Stat: The specified key does not exist' in stderr
            or 'The specified bucket does not exist' in stderr
        ):
            return Exception.__new__(ResticInitError)
        if 'unable to create lock in backend' in stderr:
            return Exception.__new__(ResticLockedError)
        index_errs = (
            'repair index',
            'rebuild-index',
            'unable to load index',
            'returned error, retrying after',
            'Load(<index',
            'Load(<lock',
        )
        if any(x in stderr for x in index_errs) or (
            stderr.startswith('Fatal:') and 'invalid data returned' in stderr
        ):
            return Exception.__new__(ResticBadIndexError)
        if proc.returncode < 0 and abs(proc.returncode) == SIGPIPE:
            # https://github.com/restic/restic/issues/1466
            # https://github.com/restic/restic/pull/2546
            return Exception.__new__(ResticConnError)
        if 'out of memory' in stderr:
            return Exception.__new__(ResticOom)
        return Exception.__new__(cls)

    @property
    def name(self) -> str:
        """Get the name of the class of the restic exception

        Returns:
            str: class name
        """
        return type(self).__name__

    def __str__(self):
        if self.returncode < 0:
            signum = abs(self.returncode)
            try:
                # https://github.com/PyCQA/pylint/issues/2804
                sig = Signals(signum).name  # pylint: disable=no-member
            except ValueError:
                sig = f'signal {signum}'
            if signum == SIGPIPE:
                # https://github.com/restic/restic/issues/1466
                return f'Timeout: received SIGPIPE. {self.stderr}'
            return f'(killed with {sig}) {self.stderr}'
        return f'(return code {self.returncode}) {self.stderr}'


class ResticOom(ResticError):
    """Raised when restic breaks due to an OOM condition"""

    __module__ = 'restic'


class ResticBadIndexError(ResticError):
    """Raised when restic raises an index corruption error"""

    __module__ = 'restic'


class ResticAccessError(ResticError):
    """Raised when S3 access/secret keys are incorrect for restic"""

    __module__ = 'restic'


class ResticInitError(ResticError):
    """Raised when a restic repo connects but is not initialized"""

    __module__ = 'restic'


class ResticConnError(ResticError):
    """Raised when a connection is refused or times out from ceph"""

    __module__ = 'restic'


class ResticLockedError(ResticError):
    """Raised when an operation on a repo fails due to a lock held on it"""

    __module__ = 'restic'

    def __init__(self, proc: Union[CalledProcessError, CompletedProcess]):
        super().__init__(proc)
        if match := LOCK_RE.search(self.stderr):
            self.pid = int(match.group(1))
            self.host = match.group(2)
        else:
            logging.warning('failed to parse PID and host from %r', self.stderr)
            self.pid = None
            self.host = None

    def unlock_ok(self) -> bool:
        """Check if a locked repo can be unlocked

        Returns:
            bool: true if restic unlock can be run safely
        """
        if self.pid is None:
            return False
        if platform.node() != self.host:
            return False
        try:
            os.kill(self.pid, 0)
        except ProcessLookupError:
            return True
        return False

    def __str__(self):
        if self.pid:
            return f"repo is in use and locked by PID {self.pid} on {self.host}"
        return super().__str__()



import warnings

from .compilers.C import msvc

__all__ = ["MSVCCompiler"]

MSVCCompiler = msvc.Compiler


def __getattr__(name):
    if name == '_get_vc_env':
        warnings.warn(
            "_get_vc_env is private; find an alternative (pypa/distutils#340)"
        )
        return msvc._get_vc_env
    raise AttributeError(name)



from typing import Any
from . import Adapter

DEFAULT_BUFFER_SIZE: int

class BuiltinSSLAdapter(Adapter):
    CERT_KEY_TO_ENV: Any
    CERT_KEY_TO_LDAP_CODE: Any
    def __init__(self, certificate, private_key, certificate_chain: Any | None = ..., ciphers: Any | None = ...) -> None: ...
    @property
    def context(self): ...
    @context.setter
    def context(self, context) -> None: ...
    def bind(self, sock): ...
    def wrap(self, sock): ...
    def get_environ(self, sock): ...
    def makefile(self, sock, mode: str = ..., bufsize: int = ...): ...



import logging

from pyroute2.ndb.main import NDB

log = logging.getLogger(__name__)


class ObjectProxy(dict):
    def __init__(self, obj):
        self._obj = obj

    def __getattribute__(self, key):
        if key[:4] == 'set_':

            def set_value(value):
                self[key[4:]] = value
                return self

            return set_value
        try:
            return self[key]
        except KeyError:
            return super(ObjectProxy, self).__getattribute__(key)

    def __setattr__(self, key, value):
        if key == '_obj':
            super(ObjectProxy, self).__setattr__(key, value)
        else:
            super(ObjectProxy, self).__getattribute__('_obj')[key] = value

    def __getitem__(self, key):
        return super(ObjectProxy, self).__getattribute__('_obj')[key]

    def __setitem__(self, key, value):
        super(ObjectProxy, self).__getattribute__('_obj')[key] = value

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if hasattr(self, 'commit'):
            self.commit()

    def __repr__(self):
        return repr(super(ObjectProxy, self).__getattribute__('_obj'))

    def __contains__(self, key):
        return key in super(ObjectProxy, self).__getattribute__('_obj')

    def get_ndb_object(self):
        return self._obj

    def keys(self):
        return self._obj.keys()

    def items(self):
        return self._obj.items()

    def values(self):
        return self._obj.values()

    def __iter__(self):
        return self._obj.__iter__()

    @property
    def _mode(self):
        return 'implicit'


class Interface(ObjectProxy):
    def add_ip(self, *argv, **kwarg):
        self._obj.add_ip(*argv, **kwarg)
        return self

    def del_ip(self, *argv, **kwarg):
        self._obj.del_ip(*argv, **kwarg)
        return self

    def add_port(self, *argv, **kwarg):
        self._obj.add_port(*argv, **kwarg)
        return self

    def del_port(self, *argv, **kwarg):
        self._obj.del_port(*argv, **kwarg)
        return self

    def commit(self, *argv, **kwarg):
        self._obj.commit(*argv, **kwarg)
        return self

    def up(self):
        self._obj.set('state', 'up')
        return self

    def down(self):
        self._obj.set('state', 'down')
        return self

    def remove(self):
        self._obj.remove()
        return self

    @property
    def if_master(self):
        return self._obj.get('master', None)

    @property
    def ipaddr(self):
        return tuple(self._obj.ipaddr.dump().select('address', 'prefixlen'))


class Interfaces(ObjectProxy):
    text_create = '''
When `create().commit()` fails, the failed interface object behaves
differently in IPDB and NDB. IPDB saves the failed object in the database,
while the NDB database contains only the system reflection, and the failed
object may stay only being referenced by a variable.

`KeyError: 'object exists'` vs. `CreateException`
'''

    def __getitem__(self, key):
        return Interface(super(Interfaces, self).__getitem__(key))

    def __iter__(self):
        return iter(self.keys())

    def add(self, *argv, **kwarg):
        return self.create(*argv, **kwarg)

    def create(self, *argv, **kwarg):
        log.warning(self.text_create)
        return Interface(self._obj.create(*argv, **kwarg))

    def keys(self):
        ret = []
        for record in self._obj.dump():
            ret += [record.ifname, record.index]
        return ret

    def has_key(self, key):
        return key in self.keys()


class NoIPDB(object):
    text_create = '''
IPDB has a shortcut method to create interfaces: `ipdb.create(...)`.

NDB has `create()` methods only under respective views:
`ndb.interfaces.create(...)`, `ndb.addresses.create(...)` etc.
'''

    text_nl = '''
Unlike IPDB, NDB can work with many netlink sources. The default one
referenced as `localhost`::

    #
    # these two statements are equivalent:
    #
    ndb.sources['localhost'].nl.get_links()
    ipdb.nl.get_links()

'''

    def __init__(self, *argv, **kwarg):
        if argv or kwarg:
            log.warning(
                '%s does not support IPDB parameters, ignoring',
                self.__class__.__name__,
            )
        if len(argv) > 0 or 'nl' in kwarg:
            log.warning(
                '%s does not support shared netlink sources,'
                ' ignoring `nl` and starting with local IPRoute',
                self.__class__.__name__,
            )

        self._ndb = NDB()
        self.interfaces = Interfaces(self._ndb.interfaces)

    @property
    def nl(self):
        log.warning(self.text_nl)
        return self._ndb.sources['localhost'].nl

    @property
    def ipaddr(self):
        ret = dict([(x.index, []) for x in self._ndb.interfaces.dump()])
        for record in self._ndb.addresses.dump():
            ret[record.index].append((record.address, record.prefixlen))
        return ret

    def create(self, *argv, **kwarg):
        log.warning(self.text_create)
        return self.interfaces.create(*argv, **kwarg)

    def release(self):
        self._ndb.close()



"""
Module for managing quotas on POSIX-like systems.
"""

import logging

import salt.utils.path
import salt.utils.platform
from salt.exceptions import CommandExecutionError, SaltInvocationError

log = logging.getLogger(__name__)


# Define a function alias in order not to shadow built-in's
__func_alias__ = {"set_": "set"}


def __virtual__():
    """
    Only work on POSIX-like systems with setquota binary available
    """
    if not salt.utils.platform.is_windows() and salt.utils.path.which("setquota"):
        return "quota"
    return (
        False,
        "The quota execution module cannot be loaded: the module is only "
        "available on POSIX-like systems with the setquota binary available.",
    )


def report(mount):
    """
    Report on quotas for a specific volume

    CLI Example:

    .. code-block:: bash

        salt '*' quota.report /media/data
    """
    ret = {mount: {}}
    ret[mount]["User Quotas"] = _parse_quota(mount, "-u")
    ret[mount]["Group Quotas"] = _parse_quota(mount, "-g")
    return ret


def _parse_quota(mount, opts):
    """
    Parse the output from repquota. Requires that -u -g are passed in
    """
    cmd = f"repquota -vp {opts} {mount}"
    out = __salt__["cmd.run"](cmd, python_shell=False).splitlines()
    mode = "header"

    if "-u" in opts:
        quotatype = "Users"
    elif "-g" in opts:
        quotatype = "Groups"
    ret = {quotatype: {}}

    for line in out:
        if not line:
            continue
        comps = line.split()
        if mode == "header":
            if "Block grace time" in line:
                blockg, inodeg = line.split(";")
                blockgc = blockg.split(": ")
                inodegc = inodeg.split(": ")
                ret["Block Grace Time"] = blockgc[-1:]
                ret["Inode Grace Time"] = inodegc[-1:]
            elif line.startswith("-"):
                mode = "quotas"
        elif mode == "quotas":
            if len(comps) < 8:
                continue
            if not comps[0] in ret[quotatype]:
                ret[quotatype][comps[0]] = {}
            ret[quotatype][comps[0]]["block-used"] = comps[2]
            ret[quotatype][comps[0]]["block-soft-limit"] = comps[3]
            ret[quotatype][comps[0]]["block-hard-limit"] = comps[4]
            ret[quotatype][comps[0]]["block-grace"] = comps[5]
            ret[quotatype][comps[0]]["file-used"] = comps[6]
            ret[quotatype][comps[0]]["file-soft-limit"] = comps[7]
            ret[quotatype][comps[0]]["file-hard-limit"] = comps[8]
            ret[quotatype][comps[0]]["file-grace"] = comps[9]
    return ret


def set_(device, **kwargs):
    """
    Calls out to setquota, for a specific user or group

    CLI Example:

    .. code-block:: bash

        salt '*' quota.set /media/data user=larry block-soft-limit=1048576
        salt '*' quota.set /media/data group=painters file-hard-limit=1000
    """
    empty = {
        "block-soft-limit": 0,
        "block-hard-limit": 0,
        "file-soft-limit": 0,
        "file-hard-limit": 0,
    }

    current = ret = None
    cmd = "setquota"
    if "user" in kwargs:
        cmd += " -u {} ".format(kwargs["user"])
        parsed = _parse_quota(device, "-u")
        if kwargs["user"] in parsed:
            current = parsed["Users"][kwargs["user"]]
        else:
            current = empty
        ret = "User: {}".format(kwargs["user"])

    if "group" in kwargs:
        if "user" in kwargs:
            raise SaltInvocationError("Please specify a user or group, not both.")
        cmd += " -g {} ".format(kwargs["group"])
        parsed = _parse_quota(device, "-g")
        if kwargs["group"] in parsed:
            current = parsed["Groups"][kwargs["group"]]
        else:
            current = empty
        ret = "Group: {}".format(kwargs["group"])

    if not current:
        raise CommandExecutionError("A valid user or group was not found")

    for limit in (
        "block-soft-limit",
        "block-hard-limit",
        "file-soft-limit",
        "file-hard-limit",
    ):
        if limit in kwargs:
            current[limit] = kwargs[limit]

    cmd += "{} {} {} {} {}".format(
        current["block-soft-limit"],
        current["block-hard-limit"],
        current["file-soft-limit"],
        current["file-hard-limit"],
        device,
    )

    result = __salt__["cmd.run_all"](cmd, python_shell=False)
    if result["retcode"] != 0:
        raise CommandExecutionError(
            "Unable to set desired quota. Error follows: \n{}".format(result["stderr"])
        )
    return {ret: current}


def warn():
    """
    Runs the warnquota command, to send warning emails to users who
    are over their quota limit.

    CLI Example:

    .. code-block:: bash

        salt '*' quota.warn
    """
    __salt__["cmd.run"]("quotawarn")


def stats():
    """
    Runs the quotastats command, and returns the parsed output

    CLI Example:

    .. code-block:: bash

        salt '*' quota.stats
    """
    ret = {}
    out = __salt__["cmd.run"]("quotastats").splitlines()
    for line in out:
        if not line:
            continue
        comps = line.split(": ")
        ret[comps[0]] = comps[1]

    return ret


def on(device):
    """
    Turns on the quota system

    CLI Example:

    .. code-block:: bash

        salt '*' quota.on
    """
    cmd = f"quotaon {device}"
    __salt__["cmd.run"](cmd, python_shell=False)
    return True


def off(device):
    """
    Turns off the quota system

    CLI Example:

    .. code-block:: bash

        salt '*' quota.off
    """
    cmd = f"quotaoff {device}"
    __salt__["cmd.run"](cmd, python_shell=False)
    return True


def get_mode(device):
    """
    Report whether the quota system for this device is on or off

    CLI Example:

    .. code-block:: bash

        salt '*' quota.get_mode
    """
    ret = {}
    cmd = f"quotaon -p {device}"
    out = __salt__["cmd.run"](cmd, python_shell=False)
    for line in out.splitlines():
        comps = line.strip().split()
        if comps[3] not in ret:
            if comps[0].startswith("quotaon"):
                if comps[1].startswith("Mountpoint"):
                    ret[comps[4]] = "disabled"
                    continue
                elif comps[1].startswith("Cannot"):
                    ret[device] = "Not found"
                    return ret
                continue
            ret[comps[3]] = {
                "device": comps[4].replace("(", "").replace(")", ""),
            }
        ret[comps[3]][comps[0]] = comps[6]
    return ret



"""
Migration tools
"""

import os.path
import shutil

import salt.syspaths as syspaths


def migrate_paths(opts):
    """
    Migrate old minion and master pki file paths to new ones.
    """
    oldpki_dir = os.path.join(syspaths.CONFIG_DIR, "pki")

    if not os.path.exists(oldpki_dir):
        # There's not even a pki directory, don't bother migrating
        return

    newpki_dir = opts["pki_dir"]

    if opts["default_include"].startswith("master"):
        keepers = [
            "master.pem",
            "master.pub",
            "syndic_master.pub",
            "minions",
            "minions_pre",
            "minions_rejected",
        ]
        if not os.path.exists(newpki_dir):
            os.makedirs(newpki_dir)
        for item in keepers:
            oi_path = os.path.join(oldpki_dir, item)
            ni_path = os.path.join(newpki_dir, item)
            if os.path.exists(oi_path) and not os.path.exists(ni_path):
                shutil.move(oi_path, ni_path)

    if opts["default_include"].startswith("minion"):
        keepers = [
            "minion_master.pub",
            "minion.pem",
            "minion.pub",
        ]
        if not os.path.exists(newpki_dir):
            os.makedirs(newpki_dir)
        for item in keepers:
            oi_path = os.path.join(oldpki_dir, item)
            ni_path = os.path.join(newpki_dir, item)
            if os.path.exists(oi_path) and not os.path.exists(ni_path):
                shutil.move(oi_path, ni_path)



"""
The local returner is used to test the returner interface, it just prints the
return data to the console to verify that it is being passed properly

To use the local returner, append '--return local' to the salt command. ex:

.. code-block:: bash

    salt '*' test.ping --return local
"""


def returner(ret):
    """
    Print the return data to the terminal to verify functionality
    """
    print(ret)


def event_return(event):
    """
    Print event return data to the terminal to verify functionality
    """

    print(event)



import datetime
import re
import string

from salt.exceptions import InvalidConfigError, SaltInvocationError

SALT_RUNTYPE_MASTER = 0
SALT_RUNTYPE_MASTER_IMPERSONATING = 1
SALT_RUNTYPE_MASTER_PEER_RUN = 2
SALT_RUNTYPE_MINION_LOCAL = 3
SALT_RUNTYPE_MINION_REMOTE = 4


def _get_salt_run_type(opts):
    if "vault" in opts and opts.get("__role", "minion") == "master":
        if opts.get("minion_id"):
            return SALT_RUNTYPE_MASTER_IMPERSONATING
        if "grains" in opts and "id" in opts["grains"]:
            return SALT_RUNTYPE_MASTER_PEER_RUN
        return SALT_RUNTYPE_MASTER

    config_location = opts.get("vault", {}).get("config_location")
    if config_location and config_location not in ["local", "master"]:
        raise InvalidConfigError(
            "Invalid vault configuration: config_location must be either local or master"
        )

    if config_location == "master":
        pass
    elif any(
        (
            opts.get("local", None),
            opts.get("file_client", None) == "local",
            opts.get("master_type", None) == "disable",
            config_location == "local",
        )
    ):
        return SALT_RUNTYPE_MINION_LOCAL
    return SALT_RUNTYPE_MINION_REMOTE


def iso_to_timestamp(iso_time):
    """
    Most endpoints respond with RFC3339-formatted strings
    This is a hacky way to use inbuilt tools only for converting
    to a timestamp
    """
    # drop subsecond precision to make it easier on us
    # (length would need to be 3, 6 or 9)
    iso_time = re.sub(r"\.[\d]+", "", iso_time)
    iso_time = re.sub(r"Z$", "+00:00", iso_time)
    try:
        # Python >=v3.7
        return int(datetime.datetime.fromisoformat(iso_time).timestamp())
    except AttributeError:
        # Python < v3.7
        dstr, tstr = iso_time.split("T")
        year = int(dstr[:4])
        month = int(dstr[5:7])
        day = int(dstr[8:10])
        hour = int(tstr[:2])
        minute = int(tstr[3:5])
        second = int(tstr[6:8])
        tz_pos = (tstr.find("-") + 1 or tstr.find("+") + 1) - 1
        tz_hour = int(tstr[tz_pos + 1 : tz_pos + 3])
        tz_minute = int(tstr[tz_pos + 4 : tz_pos + 6])
        if all(x == 0 for x in (tz_hour, tz_minute)):
            tz = datetime.timezone.utc
        else:
            tz_sign = -1 if tstr[tz_pos] == "-" else 1
            td = datetime.timedelta(hours=tz_hour, minutes=tz_minute)
            tz = datetime.timezone(tz_sign * td)
        return int(
            datetime.datetime(year, month, day, hour, minute, second, 0, tz).timestamp()
        )


def expand_pattern_lists(pattern, **mappings):
    """
    Expands the pattern for any list-valued mappings, such that for any list of
    length N in the mappings present in the pattern, N copies of the pattern are
    returned, each with an element of the list substituted.

    pattern:
        A pattern to expand, for example ``by-role/{grains[roles]}``

    mappings:
        A dictionary of variables that can be expanded into the pattern.

    Example: Given the pattern `` by-role/{grains[roles]}`` and the below grains

    .. code-block:: yaml

        grains:
            roles:
                - web
                - database

    This function will expand into two patterns,
    ``[by-role/web, by-role/database]``.

    Note that this method does not expand any non-list patterns.
    """
    expanded_patterns = []
    f = string.Formatter()

    # This function uses a string.Formatter to get all the formatting tokens from
    # the pattern, then recursively replaces tokens whose expanded value is a
    # list. For a list with N items, it will create N new pattern strings and
    # then continue with the next token. In practice this is expected to not be
    # very expensive, since patterns will typically involve a handful of lists at
    # most.

    for _, field_name, _, _ in f.parse(pattern):
        if field_name is None:
            continue
        (value, _) = f.get_field(field_name, None, mappings)
        if isinstance(value, list):
            token = f"{{{field_name}}}"
            expanded = [pattern.replace(token, str(elem)) for elem in value]
            for expanded_item in expanded:
                result = expand_pattern_lists(expanded_item, **mappings)
                expanded_patterns += result
            return expanded_patterns
    return [pattern]


def timestring_map(val):
    """
    Turn a time string (like ``60m``) into a float with seconds as a unit.
    """
    if val is None:
        return val
    if isinstance(val, (int, float)):
        return float(val)
    try:
        return float(val)
    except ValueError:
        pass
    if not isinstance(val, str):
        raise SaltInvocationError("Expected integer or time string")
    if not re.match(r"^\d+(?:\.\d+)?[smhd]$", val):
        raise SaltInvocationError(f"Invalid time string format: {val}")
    raw, unit = float(val[:-1]), val[-1]
    if unit == "s":
        return raw
    raw *= 60
    if unit == "m":
        return raw
    raw *= 60
    if unit == "h":
        return raw
    raw *= 24
    if unit == "d":
        return raw
    raise RuntimeError("This path should not have been hit")



"""
Microsoft Updates (KB) Management

This module provides the ability to enforce KB installations from files (.msu),
without WSUS or Windows Update

.. versionadded:: 2018.3.4
"""

import logging

import salt.utils.platform
import salt.utils.url
from salt.exceptions import CommandExecutionError, SaltInvocationError

log = logging.getLogger(__name__)


# Define the module's virtual name
__virtualname__ = "wusa"


def __virtual__():
    """
    Load only on Windows
    """
    if not salt.utils.platform.is_windows():
        return False, "Only available on Windows systems"

    return __virtualname__


def installed(name, source):
    """
    Ensure an update is installed on the minion

    Args:

        name(str):
            Name of the Windows KB ("KB123456")

        source (str):
            Source of .msu file corresponding to the KB

    Example:

    .. code-block:: yaml

        KB123456:
          wusa.installed:
            - source: salt://kb123456.msu
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    # Input validation
    if not name:
        raise SaltInvocationError('Must specify a KB "name"')
    if not source:
        raise SaltInvocationError('Must specify a "source" file to install')

    # Is the KB already installed
    if __salt__["wusa.is_installed"](name):
        ret["result"] = True
        ret["comment"] = f"{name} already installed"
        return ret

    # Check for test=True
    if __opts__["test"] is True:
        ret["result"] = None
        ret["comment"] = f"{name} would be installed"
        ret["result"] = None
        return ret

    # Cache the file
    cached_source_path = __salt__["cp.cache_file"](path=source, saltenv=__env__)
    if not cached_source_path:
        msg = 'Unable to cache {} from saltenv "{}"'.format(
            salt.utils.url.redact_http_basic_auth(source), __env__
        )
        ret["comment"] = msg
        return ret

    # Install the KB

    additional_comment = ""

    try:
        __salt__["wusa.install"](cached_source_path)
    except CommandExecutionError as exc:
        additional_comment = exc.message

    # Verify successful install
    if __salt__["wusa.is_installed"](name):
        ret["comment"] = f"{name} was installed. {additional_comment}"
        ret["changes"] = {"old": False, "new": True}
        ret["result"] = True
    else:
        ret["comment"] = f"{name} failed to install. {additional_comment}"

    return ret


def uninstalled(name):
    """
    Ensure an update is uninstalled from the minion

    Args:

        name(str):
            Name of the Windows KB ("KB123456")

    Example:

    .. code-block:: yaml

        KB123456:
          wusa.uninstalled
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    # Is the KB already uninstalled
    if not __salt__["wusa.is_installed"](name):
        ret["result"] = True
        ret["comment"] = f"{name} already uninstalled"
        return ret

    # Check for test=True
    if __opts__["test"] is True:
        ret["result"] = None
        ret["comment"] = f"{name} would be uninstalled"
        ret["result"] = None
        return ret

    # Uninstall the KB
    __salt__["wusa.uninstall"](name)

    # Verify successful uninstall
    if not __salt__["wusa.is_installed"](name):
        ret["comment"] = f"{name} was uninstalled"
        ret["changes"] = {"old": True, "new": False}
        ret["result"] = True
    else:
        ret["comment"] = f"{name} failed to uninstall"

    return ret



"""
Kapacitor state module.

:configuration: This module accepts connection configuration details either as
    parameters or as configuration settings in /etc/salt/minion on the relevant
    minions::

        kapacitor.unsafe_ssl: 'false'
        kapacitor.protocol: 'http'
        kapacitor.host: 'localhost'
        kapacitor.port: 9092

    This data can also be passed into pillar. Options passed into opts will
    overwrite options passed into pillar.

.. versionadded:: 2016.11.0
"""

import difflib

import salt.utils.files
import salt.utils.stringutils


def __virtual__():
    if "kapacitor.version" in __salt__:
        return "kapacitor"
    return (False, "kapacitor module could not be loaded")


def task_present(
    name,
    tick_script,
    task_type="stream",
    database=None,
    retention_policy="default",
    enable=True,
    dbrps=None,
):
    """
    Ensure that a task is present and up-to-date in Kapacitor.

    name
        Name of the task.

    tick_script
        Path to the TICK script for the task. Can be a salt:// source.

    task_type
        Task type. Defaults to 'stream'

    dbrps
        A list of databases and retention policies in "dbname"."rpname" format
        to fetch data from. For backward compatibility, the value of
        'database' and 'retention_policy' will be merged as part of dbrps.

        .. versionadded:: 2019.2.0

    database
        Which database to fetch data from. Defaults to None, which will use the
        default database in InfluxDB.

    retention_policy
        Which retention policy to fetch data from. Defaults to 'default'.

    enable
        Whether to enable the task or not. Defaults to True.
    """
    comments = []
    changes = []
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    task = __salt__["kapacitor.get_task"](name)
    old_script = task["script"] if task else ""
    if not dbrps:
        dbrps = []
    if database and retention_policy:
        dbrp = f"{database}.{retention_policy}"
        dbrps.append(dbrp)
    task_dbrps = [
        {"db": dbrp[0], "rp": dbrp[1]} for dbrp in (dbrp.split(".") for dbrp in dbrps)
    ]

    if tick_script.startswith("salt://"):
        script_path = __salt__["cp.cache_file"](tick_script, __env__)
    else:
        script_path = tick_script

    with salt.utils.files.fopen(script_path, "r") as file:
        new_script = salt.utils.stringutils.to_unicode(file.read()).replace(
            "\t", "    "
        )

    is_up_to_date = task and (
        old_script == new_script
        and task_type == task["type"]
        and task["dbrps"] == task_dbrps
    )

    if is_up_to_date:
        comments.append("Task script is already up-to-date")
    else:
        if __opts__["test"]:
            ret["result"] = None
            comments.append("Task would have been updated")
        else:
            result = __salt__["kapacitor.define_task"](
                name,
                script_path,
                task_type=task_type,
                database=database,
                retention_policy=retention_policy,
                dbrps=dbrps,
            )
            ret["result"] = result["success"]
            if not ret["result"]:
                comments.append("Could not define task")
                if result.get("stderr"):
                    comments.append(result["stderr"])
                ret["comment"] = "\n".join(comments)
                return ret

        if old_script != new_script:
            ret["changes"]["TICKscript diff"] = "\n".join(
                difflib.unified_diff(
                    old_script.splitlines(),
                    new_script.splitlines(),
                )
            )
            comments.append("Task script updated")

        if not task or task["type"] != task_type:
            ret["changes"]["type"] = task_type
            comments.append("Task type updated")

        if not task or task["dbrps"] != task_dbrps:
            ret["changes"]["dbrps"] = task_dbrps
            comments.append("Task dbrps updated")

    if enable:
        if task and task["enabled"]:
            comments.append("Task is already enabled")
        else:
            if __opts__["test"]:
                ret["result"] = None
                comments.append("Task would have been enabled")
            else:
                result = __salt__["kapacitor.enable_task"](name)
                ret["result"] = result["success"]
                if not ret["result"]:
                    comments.append("Could not enable task")
                    if result.get("stderr"):
                        comments.append(result["stderr"])
                    ret["comment"] = "\n".join(comments)
                    return ret
                comments.append("Task was enabled")
            ret["changes"]["enabled"] = {"old": False, "new": True}
    else:
        if task and not task["enabled"]:
            comments.append("Task is already disabled")
        else:
            if __opts__["test"]:
                ret["result"] = None
                comments.append("Task would have been disabled")
            else:
                result = __salt__["kapacitor.disable_task"](name)
                ret["result"] = result["success"]
                if not ret["result"]:
                    comments.append("Could not disable task")
                    if result.get("stderr"):
                        comments.append(result["stderr"])
                    ret["comment"] = "\n".join(comments)
                    return ret
                comments.append("Task was disabled")
            ret["changes"]["enabled"] = {"old": True, "new": False}

    ret["comment"] = "\n".join(comments)
    return ret


def task_absent(name):
    """
    Ensure that a task is absent from Kapacitor.

    name
        Name of the task.
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    task = __salt__["kapacitor.get_task"](name)

    if task:
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = "Task would have been deleted"
        else:
            result = __salt__["kapacitor.delete_task"](name)
            ret["result"] = result["success"]
            if not ret["result"]:
                ret["comment"] = "Could not disable task"
                if result.get("stderr"):
                    ret["comment"] += "\n" + result["stderr"]
                return ret
            ret["comment"] = "Task was deleted"
        ret["changes"][name] = "deleted"
    else:
        ret["comment"] = "Task does not exist"

    return ret



"""
Display return data as a progress bar
"""

try:
    import progressbar

    HAS_PROGRESSBAR = True
except ImportError:
    HAS_PROGRESSBAR = False


def __virtual__():
    return True if HAS_PROGRESSBAR else False


def output(ret, bar, **kwargs):  # pylint: disable=unused-argument
    """
    Update the progress bar
    """
    if "return_count" in ret:
        val = ret["return_count"]
        # Avoid to fail if targets are behind a syndic. In this case actual return count will be
        # higher than targeted by MoM itself.
        # TODO: implement a way to get the proper target minions count and remove this workaround.
        # Details are in #44239.
        if val > bar.maxval:
            bar.maxval = val
        bar.update(val)
    return ""


def progress_iter(progress):
    """
    Initialize and return a progress bar iter
    """
    widgets = [
        progressbar.Percentage(),
        " ",
        progressbar.Bar(),
        " ",
        progressbar.Timer(),
        " Returns: [",
        progressbar.Counter(),
        "/{}]".format(progress["minion_count"]),
    ]
    bar = progressbar.ProgressBar(widgets=widgets, maxval=progress["minion_count"])
    bar.start()
    return bar



from pyroute2.protocols import msg


class icmpmsg(msg):
    fields = [('type', 'uint8'), ('code', 'uint8'), ('csum', 'be32')]


class icmp_router_adv(icmpmsg):
    fields = icmpmsg.fields + [
        ('addrs_num', 'uint8'),
        ('alen', 'uint8'),
        ('lifetime', 'be32'),
        ('addrs', 'routers'),
    ]
